# -*- coding: utf-8 -*-
import logging
from typing import List

import numpy as np
from scipy.signal import find_peaks
from scipy.optimize import curve_fit

from PyQt5.QtGui import QColor

from .basic_widgets import (BasicInputParametersWidget, ConfirmButton,
                            RoundedPushButton, Smooth1DPlot)
from .signal_connection import SignalConnector, SignalContainer
from .roi.roi_widgets import Roi1D
from .roi.roi_containers import BasicROIContainer

from ..config import read_config
from ..utils import Icon, RoiParameters, show_error

logger = logging.getLogger(__name__)


class RadialProfileWidget(BasicROIContainer, Smooth1DPlot):
    _DefaultRoiWidth = 50
    _DefaultNewRoiParameters = dict(radius=10, width=5)

    def __init__(self, signal_connector: SignalConnector,
                 parent=None):
        BasicROIContainer.__init__(self, signal_connector)
        Smooth1DPlot.__init__(self, parent)
        self._peaks_setup = None
        self._fit_parameters_dict = read_config(PeaksSetupWindow.NAME)
        self.update_image()

    def __init_toolbars__(self):
        super().__init_toolbars__()

        fit_toolbar = self.addToolBar('Fitting')
        fit_toolbar.setStyleSheet('background-color: black;')

        find_peaks_widget = ConfirmButton(Icon('find'), text='Find peaks?')
        find_peaks_widget.label_widget.setStyleSheet(
            'QLabel { color : white ; }')
        find_peaks_widget.clicked.connect(self.find_peaks)
        fit_toolbar.addWidget(find_peaks_widget)

        fit_peaks_widget = ConfirmButton(Icon('fit'), text='Fit selected peaks?')
        fit_peaks_widget.label_widget.setStyleSheet(
            'QLabel { color : white ; }')
        fit_peaks_widget.clicked.connect(self.fit_selected)
        fit_toolbar.addWidget(fit_peaks_widget)

        setup_action = fit_toolbar.addAction(Icon('setup'), 'Fit setup')
        setup_action.triggered.connect(self.open_peaks_setup)

        segments_toolbar = self.addToolBar('Segments')
        segments_toolbar.setStyleSheet('background-color: black;')

        create_roi_widget = RoundedPushButton(
            icon=Icon('add'), radius=30)
        create_roi_widget.clicked.connect(self.emit_create_segment)
        segments_toolbar.addWidget(create_roi_widget)

        delete_selected_widget = ConfirmButton(
            Icon('delete'), text='Delete selected roi?')
        delete_selected_widget.clicked.connect(self.emit_delete_selected_roi)
        delete_selected_widget.label_widget.setStyleSheet(
            'QLabel { color : white ; }')

        segments_toolbar.addWidget(delete_selected_widget)
        fix_all = RoundedPushButton(icon=Icon('fix_all'), radius=120, background_color=QColor(0, 0, 0, 0))
        fix_all.setFixedWidth(60)
        fix_all.setFixedHeight(30)
        fix_all.clicked.connect(self.fix_all)
        segments_toolbar.addWidget(fix_all)
        unfix_all = RoundedPushButton(icon=Icon('unfix_all'), radius=120, background_color=QColor(0, 0, 0, 0))
        unfix_all.setFixedWidth(60)
        unfix_all.setFixedHeight(30)
        unfix_all.clicked.connect(self.unfix_all)
        segments_toolbar.addWidget(unfix_all)

    def process_signal(self, s: SignalContainer):
        update_image = False
        for _ in s.image_changed():
            update_image = True
        for _ in s.geometry_changed():
            update_image = True
        BasicROIContainer.process_signal(self, s)
        if update_image:
            self.update_image()

    def _on_scale_changed(self):
        self.update_x_axis()

    def find_peaks(self):
        # TODO show message if number of peaks exceeds max number and suggest to increase sigma.
        if self._fit_parameters_dict.get('sigma_find', None) is not None:
            self.set_sigma(self._fit_parameters_dict['sigma_find'])
        peaks = find_peaks(self.smoothed_y)[0]
        sc = SignalContainer(app_node=self)
        for i, peak in enumerate(peaks):
            segment = RoiParameters(peak * self.image.scale, self._DefaultRoiWidth * self.image.scale,
                                    name=f'Proposed ring {i}')
            sc.segment_created(segment)
        sc.send()

    def fit_selected(self):
        if self._fit_parameters_dict.get('sigma_fit', None) is not None:
            self.set_sigma(self._fit_parameters_dict['sigma_fit'])
        sc = SignalContainer(app_node=self)
        fit_params = FitParameters(self.x, self.smoothed_y, self.image.scale)
        for value in self.get_selected():
            fit_params.add_value(value)
            value = list(fit_params.fit())[0]
            if value is not None:
                sc.segment_moved(value, signal_type='broadcast')
                sc.segment_fixed(value)
            fit_params.clear()
        sc.send()

    def fit_together(self):
        if self._fit_parameters_dict.get('sigma_fit', None) is not None:
            self.set_sigma(self._fit_parameters_dict['sigma_fit'])
        sc = SignalContainer(app_node=self)
        fit_params = FitParameters(self.x, self.smoothed_y, self.image.scale)
        fit_params.add_values(self.get_selected())
        for value in fit_params.fit():
            sc.segment_moved(value, signal_type='broadcast')
            sc.segment_fixed(value)
        sc.send()

    def _get_default_roi_parameters(self):
        scale = self.image.scale
        param_dict = self._DefaultNewRoiParameters.copy()
        param_dict['radius'] *= scale
        param_dict['width'] *= scale
        return RoiParameters(**param_dict)

    def open_peaks_setup(self):
        self._peaks_setup = PeaksSetupWindow()
        self._peaks_setup.apply_signal.connect(self.set_fit_parameters)
        self._peaks_setup.close_signal.connect(self.close_peaks_setup)
        self._peaks_setup.show()

    def set_fit_parameters(self, params: dict):
        self._fit_parameters_dict = params

    def close_peaks_setup(self):
        self._peaks_setup = None

    def emit_create_segment(self, *args):
        BasicROIContainer.emit_create_segment(
            self, self._get_default_roi_parameters()
        )

    def emit_delete_selected_roi(self):
        sc = SignalContainer()
        for roi in [roi for roi in self.roi_dict.values()
                    if roi.active]:
            sc.segment_deleted(roi)
        self.signal_connector.emit_upward(sc)

    def _get_roi(self, params: RoiParameters):
        return Roi1D(params)

    def _add_item(self, roi):
        self.image_view.plot_item.addItem(roi)

    def _remove_item(self, roi):
        self.image_view.plot_item.removeItem(roi)

    def update_image(self):
        if self.image.rr is None or self.image.image is None:
            return
        self.y = get_radial_profile(self.image.image, self.image.rr)
        self.update_x_axis()

    def update_x_axis(self, plot: bool = True):
        rr = self.image.rr
        self._x = np.linspace(rr.min(), rr.max(), self.smoothed_y.size) * self.image.scale
        if plot:
            self.plot()


class PeaksSetupWindow(BasicInputParametersWidget):
    P = BasicInputParametersWidget.InputParameters

    PARAMETER_TYPES = (P('max_peaks_number',
                         'Maximum number of peaks',
                         int, 'Do not recommended to put high numbers'),
                       P('init_width', 'Peaks width', float,
                         'Gaussian fitting will start with this number'),
                       P('sigma_find', 'Sigma to find peaks', float,
                         'Default sigma value for gaussian smooth\n'
                         'applied before initial peaks finding to\n'
                         'avoid noise peaks. To use current lambda, \n'
                         'leave empty.', True),
                       P('sigma_fit', 'Sigma to find peaks', float,
                         'Default sigma value for gaussian smooth\n'
                         'applied before gaussian fitting of \n'
                         'found peaks. To use current lambda, \n'
                         'leave empty.', True)
                       )

    NAME = 'Fitting parameters'


class FitParameters(object):
    _MAXIMUM_NUMBER_OF_PEAKS = 6

    @property
    def x(self):
        if self._x1 is None or self._x2 is None:
            return self._x
        else:
            return self._x[self._x1:self._x2]

    @property
    def y(self):
        if self._x1 is None or self._x2 is None:
            return self._y
        else:
            return self._y[self._x1:self._x2]

    @property
    def bounds(self):
        return tuple(self._lower_bounds), tuple(self._upper_bounds)

    @property
    def init_parameters(self):
        return tuple(self._init_conditions)

    def __init__(self, x, y, scale: float):
        self._x = x
        self._y = y
        self._scale = scale
        self._number_of_rois = 0
        self._upper_bounds = []
        self._lower_bounds = []
        self._init_conditions = []
        self._values = []
        self._x1 = None
        self._x2 = None

    def add_value(self, value: RoiParameters):
        mu_min, mu_max = value.radius - value.width / 2, value.radius + value.width / 2
        x1, x2 = (int(mu_min / self._scale),
                  int(mu_max / self._scale))
        data = self._y[x1:x2]
        if not data.size:
            return
        A = data.max()
        sigma = value.width / 2
        mu = value.radius
        B = 0
        self._init_conditions.extend([A, mu, sigma, B])
        self._lower_bounds.extend([0, mu_min, 0, 0])
        self._upper_bounds.extend([A, mu_max, sigma * 2, A])

        if self._x1 is None or self._x1 > x1:
            self._x1 = x1
        if self._x2 is None or self._x2 < x2:
            self._x2 = x2
        self._values.append(value)
        self._number_of_rois += 1

    def add_values(self, values: List[RoiParameters]):
        for value in values:
            self.add_value(value)

    def fit(self):
        if not self._number_of_rois:
            return ()
        elif self._number_of_rois > self._MAXIMUM_NUMBER_OF_PEAKS:
            show_error(f'The maximum number of peaks for this option is '
                       f'limited ({self._MAXIMUM_NUMBER_OF_PEAKS}). "Fit together" '
                       f'option is only necessary for overlapping peaks.',
                       'Maximum number of peaks exceeded')
            return ()
        elif self._number_of_rois == 1:
            func = gauss
        else:
            func = multi_gauss
        try:
            res = curve_fit(func, self.x, self.y, self.init_parameters,
                            bounds=self.bounds)
            parameters = res[0]
            for i, value in enumerate(self._values):
                A, mu, sigma, B = parameters[4 * i:(4 * i + 4)]
                yield value._replace(radius=mu, width=sigma * 2,
                                     fit_r_parameters=(A, mu, sigma, B))
        except RuntimeError:
            return ()

    def clear(self):
        self._number_of_rois = 0
        self._upper_bounds = []
        self._lower_bounds = []
        self._init_conditions = []
        self._values = []
        self._x1 = None
        self._x2 = None


def gauss(x, *p):
    A, mu, sigma, B = p
    return A * np.exp(-(x - mu) ** 2 / (2. * sigma ** 2)) + B


def multi_gauss(x, *p):
    if len(p) % 4:
        raise ValueError(f'Wrong number of parameters {len(p)}.')
    res = np.zeros_like(x)
    for i in range(len(p) // 4):
        res += gauss(x, *p[4 * i:(4 * i + 4)])
    return res


def get_radial_profile(img, r):
    assert img.shape == r.shape
    r = r.astype(np.int)

    tbin = np.bincount(r.ravel(), img.ravel())
    nr = np.bincount(r.ravel())
    radial_profile = np.nan_to_num(tbin / nr)
    return radial_profile



import logging
import weakref
from copy import deepcopy
from collections import defaultdict, namedtuple
from functools import wraps
from abc import abstractmethod

import numpy as np

from PyQt5.QtCore import QObject, pyqtSignal

from .exceptions import SignalNotFoundError, AppNodeNotProvidedError
from .global_context import Image
from ..utils import RoiParameters

logger = logging.getLogger(__name__)

_SIGNAL_TYPES = ('broadcast', 'onlyForNames', 'exceptForNames')
# TODO change to ENUM

_DEFAULT_SIGNAL_TYPE = 'broadcast'

_GlobalImageObject = Image()
# TODO change to dependency injection


class SignalKeys(object):
    # TODO: change to ENUM
    _image_changed_key = 'image_changed'
    _geometry_changed_key = 'geometry_changed'
    _geometry_changed_finish_key = 'geometry_changed_finish'
    _transformation_key = 'transformation'

    _status_changed_key = 'status_changed'
    _name_changed_key = 'name_changed'
    _scale_changed_key = 'scale_changed'
    _type_changed_key = 'type_changed'

    _segment_created_key = 'segment_created'
    _segment_deleted_key = 'segment_deleted'
    _segment_moved_key = 'segment_moved'

    _segment_fixed_key = 'segment_fixed'
    _segment_unfixed_key = 'segment_unfixed'
    _intensity_limits_changed_key = 'intensity_limits_changed'


_SPECIAL_SIGNAL_TYPES = {
    SignalKeys._name_changed_key: 'exceptForNames',
    SignalKeys._segment_moved_key: 'exceptForNames',
    SignalKeys._geometry_changed_key: 'exceptForNames',
    SignalKeys._type_changed_key: 'exceptForNames'
}


class BasicSignalContainer(object):
    @property
    def adding_finished(self):
        return bool(self._add_later)

    @property
    def app_node(self):
        return self._app_node() if self._app_node else None

    def __init__(self, signals: defaultdict = None, app_node: 'AppNode' = None):
        self._signals = signals or defaultdict(list)
        self._add_later = defaultdict(list)
        self._app_node = weakref.ref(app_node) if app_node else None

    def send(self, app_node=None):
        app_node = app_node or self.app_node
        if not app_node:
            raise AppNodeNotProvidedError()
        app_node.signal_connector.emit_upward(self)

    def append(self, signal: 'Signal', copy: bool = False,
               add_later: bool = False):
        if not copy:
            if add_later:
                self._add_later[signal.name].append(signal)
            else:
                self._signals[signal.name].append(signal)
        else:
            new_container = self.copy()
            new_container.append(signal, add_later=add_later)
            return new_container

    def append_later(self, signal: 'Signal'):
        self.append(signal, add_later=True)

    def finish_adding_later(self):
        for k, v in self._add_later.items():
            self._signals[k] += v
        self._add_later = defaultdict(list)

    def remove(self, signal, copy: bool = True) -> 'BasicSignalContainer' or None:
        # TODO: Add option to remove later?
        try:
            self._signals[signal.name].remove(signal)
        except ValueError:
            raise SignalNotFoundError()
        if copy:
            new_container = self.copy()
            self.append(signal)
            return new_container

    def copy(self, copy_add_later: bool = True):
        new_signals_dict = defaultdict(list)
        for signal in self:
            new_signals_dict[signal.name].append(signal.copy())
        new_container = self.__class__(new_signals_dict)
        if copy_add_later:
            for k, v in self._add_later.items():
                new_container._add_later[k] += [signal.copy() for signal in v]
        return new_container

    def __iter__(self):
        for key in self._signals.keys():
            for signal in self._signals[key]:
                yield signal

    def __getitem__(self, item):
        return self._signals[item]

    def __repr__(self):
        return '\n'.join([str(signal) for signal in self])


def _overload_signal_container_functions(func):
    @wraps(func)
    def wrapper(self, data=None, *args, **kwargs):
        key, data_class = func(self)
        if data is None:
            return self[key]
        else:
            self.add_signal(key, data_class(data),
                            *args, **kwargs)
            return self

    return wrapper


class SignalContainer(BasicSignalContainer, SignalKeys):
    def add_signal(self,
                   signal_name, data, signal_type=None,
                   address_names: list or str = None,
                   add_later: bool = False):
        if not address_names:
            address_names = []
        if isinstance(address_names, str):
            address_names = [address_names]
        if not signal_type:
            signal_type = _SPECIAL_SIGNAL_TYPES.get(
                signal_name, _DEFAULT_SIGNAL_TYPE)
        signal = Signal(data, signal_name, signal_type, address_names)

        if add_later:
            self._add_later[signal.name].append(signal)
        else:
            self.append(signal)
        return self

    # The following functions are shortcuts to simplify syntax.

    @_overload_signal_container_functions
    def image_changed(self, data=None, *args, **kwargs):
        return self._image_changed_key, EmptySignalData

    @_overload_signal_container_functions
    def geometry_changed(self, data=None, *args, **kwargs):
        return self._geometry_changed_key, EmptySignalData

    @_overload_signal_container_functions
    def geometry_changed_finish(self, data=None, *args, **kwargs):
        return self._geometry_changed_finish_key, EmptySignalData

    @_overload_signal_container_functions
    def status_changed(self, data=None, *args, **kwargs):
        return self._status_changed_key, StatusChangedSignal

    @_overload_signal_container_functions
    def segment_created(self, data=None, *args, **kwargs):
        return self._segment_created_key, SegmentSignalData

    @_overload_signal_container_functions
    def segment_moved(self, data=None, *args, **kwargs):
        return self._segment_moved_key, SegmentSignalData

    @_overload_signal_container_functions
    def segment_deleted(self, data=None, *args, **kwargs):
        return self._segment_deleted_key, SegmentSignalData

    @_overload_signal_container_functions
    def segment_fixed(self, data=None, *args, **kwargs):
        return self._segment_fixed_key, SegmentSignalData

    @_overload_signal_container_functions
    def segment_unfixed(self, data=None, *args, **kwargs):
        return self._segment_unfixed_key, SegmentSignalData

    @_overload_signal_container_functions
    def transformation_added(self, data=None, *args, **kwargs):
        return self._transformation_key, EmptySignalData

    @_overload_signal_container_functions
    def intensity_limits_changed(self, data=None, *args, **kwargs):
        return self._intensity_limits_changed_key, EmptySignalData

    @_overload_signal_container_functions
    def name_changed(self, data=None, *args, **kwargs):
        return self._name_changed_key, SegmentSignalData

    @_overload_signal_container_functions
    def scale_changed(self, data=None, *args, **kwargs):
        return self._scale_changed_key, EmptySignalData

    @_overload_signal_container_functions
    def type_changed(self, data=None, *args, **kwargs):
        return self._type_changed_key, SegmentSignalData


class BasicSignalData(object):
    def __init__(self, data):
        self._data = data

    def __call__(self):
        return self._data

    @abstractmethod
    def __eq__(self, other):
        pass

    def copy(self):
        return self.__class__(deepcopy(self()))

    def __repr__(self):
        return f'Signal data {self.__class__}: {str(self._data)}'


class EmptySignalData(BasicSignalData):
    def __init__(self, *args):
        super().__init__(None)

    def __eq__(self, other):
        return True


class ImmutableSignalData(BasicSignalData):
    def __eq__(self, other):
        return self() == other()

    def copy(self):
        return self.__class__(self())

    def __repr__(self):
        return self._data


StatusChangedContainer = namedtuple(
    'StatusChangedSignal', 'keys status change_others')

StatusChangedContainer.__new__.__defaults__ = (True,)


class StatusChangedSignal(ImmutableSignalData):
    """
    This signal should not appear more than once in
    a SignalContainer.
    """

    def __init__(self, data: StatusChangedContainer):
        ImmutableSignalData.__init__(self, data)


class SegmentSignalData(ImmutableSignalData):
    def __call__(self) -> RoiParameters:
        return self._data


class NumpySignalData(BasicSignalData):
    def __eq__(self, other):
        return np.all(self() == other())


class Signal(object):
    __slots__ = ('type', 'name', 'data', 'address_names')

    def __init__(self, data: BasicSignalData,
                 signal_name: str,
                 signal_type: str,
                 address_names: list = None):
        """
        Signal is emitted in SignalContainer through the SignalConnector.
        It contains the data needed to be sent and additional attributes.

        'signal_type' can be one of 'broadcast', 'exceptForNames', 'onlyForNames'.
        'broadcast' is translated to all the signal connectors;
        'exceptForNames' is ignored by SignalConnector with NAME in 'address_names';
        'onlyForNames' is accepted by SignalConnector with NAME in 'address_names'.

        'signal_name' is a name representing specific functionality
        of the application.

        :param data:
        :param signal_type:
        :param signal_name:
        :param address_names:
        """
        if signal_type not in _SIGNAL_TYPES:
            raise ValueError('Unknown signal type.')
        self.type = signal_type
        self.name = signal_name
        self.address_names = address_names
        self.data = data

    def add_name(self, name: str):
        self.address_names.append(name)

    def __call__(self, *args, **kwargs):
        return self.data()

    def __repr__(self):
        return f'Signal {self.name}, type = {self.type}.'

    def copy(self):
        return Signal(self.data, self.name, self.type,
                      deepcopy(self.address_names))

    def __eq__(self, other):
        if (
                self.name == other.name and
                self.address_names == other.address_names and
                self.type == other.type and
                self.data == other.data
        ):
            return True
        return False


class SignalConnector(QObject):
    downwardSignal = pyqtSignal(object)
    upwardSignal = pyqtSignal(object)

    def __init__(self, name: str = None,
                 upper_connector: 'SignalConnector' = None):
        self.NAME = name
        QObject.__init__(self)
        self.image = _GlobalImageObject
        if upper_connector:
            upper_connector.connect_downward(self)

    def get_lower_connector(self, name: str = None) -> 'SignalConnector':
        return SignalConnector(name, self)

    def pass_downward(self, s: SignalContainer) -> SignalContainer or None:
        """Redefine to add additional conditions to pass downward signal
        or to change it."""
        singals_to_remove = list()
        if not s:
            return
        for signal in s:
            if signal.type == 'broadcast':
                continue
            if (self.NAME and signal.type == 'onlyForNames' and
                    self.NAME not in signal.address_names):
                singals_to_remove.append(signal)
            if (self.NAME and signal.type == 'exceptForNames' and
                    self.NAME in signal.address_names):
                singals_to_remove.append(signal)

        for signal in singals_to_remove:
            s = s.remove(signal)
        return s

    def pass_upward(self, s: SignalContainer) -> SignalContainer or None:
        """Redefine to add additional conditions to pass upward signal
        or to change it."""
        if not s:
            return
        for signal in s:
            if self.NAME and signal.type != 'broadcast':
                signal.add_name(self.NAME)
        return s

    def connect_downward(self, lower_connector: 'SignalConnector') -> None:
        self.downwardSignal.connect(lower_connector.emit_downward)
        lower_connector.upwardSignal.connect(self.emit_upward)

    def emit_downward(self, s: SignalContainer):
        s = self.pass_downward(s)
        if s:
            self.downwardSignal.emit(s)

    def emit_upward(self, s: SignalContainer):
        s = self.pass_upward(s)
        if s:
            self.upwardSignal.emit(s)


class AppDataHolder(SignalConnector):

    def __init__(self):
        SignalConnector.__init__(self, 'AppDataHolder')
        self.segments_dict = dict()
        self.selected_keys = list()
        self._status_changed_sent = False
        # only one StatusChangedSignal can be sent in a SignalContainer

    def connect(self, func):
        self.upwardSignal.connect(func)

    def pass_downward(self, s: SignalContainer) -> SignalContainer:

        selected_keys = list()
        # These keys are the keys of created or moved
        # rois. They will form StatusChangedSignal if not empty.
        for _ in s.geometry_changed():
            self.on_geometry_changed(s)

        for _ in s.scale_changed():
            self.on_scale_changed(s)

        for signal in s.segment_deleted():
            key = signal().key
            self.segments_dict.pop(key)
            if key in self.selected_keys:
                self.selected_keys.remove(key)

        for signal in s.segment_created():
            segment = signal()
            signal.data = self.add_segment(segment)
            selected_keys.append(signal().key)
        for signal in s.segment_moved():
            segment = signal()
            self.segments_dict[segment.key] = signal()
            selected_keys.append(signal().key)

        for signal in s.segment_fixed():
            self.segments_dict[signal().key] = signal()

        for signal in s.segment_unfixed():
            self.segments_dict[signal().key] = signal()

        signals_to_remove = list()
        for signal in s.status_changed():
            data_list = self.on_status_changed(signal())
            signals_to_remove.append(signal)
            for data in data_list:
                s.status_changed(data, add_later=True)
        for signal in signals_to_remove:
            s = s.remove(signal)

        for signal in s.name_changed():
            self.segments_dict[signal().key] = signal()
        ##

        if selected_keys and not self._status_changed_sent:
            data = StatusChangedContainer(
                tuple(selected_keys), True, True
            )
            data_list = self.on_status_changed(data)
            for data in data_list:
                s.status_changed(data)

        s.finish_adding_later()
        self._status_changed_sent = False
        return s

    def on_scale_changed(self, s: SignalContainer):
        scale = self.image.scale_change
        for k, v in self.segments_dict.items():
            self.segments_dict[k] = v._replace(radius=v.radius * scale, width=v.width * scale)
            s.segment_moved(self.segments_dict[k], add_later=True)

    def on_geometry_changed(self, s: SignalContainer):
        r_angle, r_angle_std = self.image.ring_angle, self.image.ring_angle_str
        if r_angle is not None and r_angle_std is not None:
            for k, v in self.segments_dict.items():
                if (
                        v.type == RoiParameters.roi_types.ring and
                        (v.angle != r_angle or
                         v.angle_std != r_angle_std)
                ):
                    self.segments_dict[k] = v._replace(
                        angle=r_angle, angle_std=r_angle_std)
                    s.segment_moved(self.segments_dict[k], add_later=True)

    def on_status_changed(self, data: StatusChangedContainer):
        self._status_changed_sent = True
        set_inactive_keys = list()
        set_active_keys = list()
        data_list = list()

        if not data.status:
            for k in data.keys:
                if k in self.selected_keys:
                    set_inactive_keys.append(k)
        else:
            if data.change_others:
                set_inactive_keys = [k for k in self.selected_keys if
                                     k not in data.keys]
            set_active_keys = [k for k in data.keys if
                               k not in self.selected_keys]
            self.selected_keys += set_active_keys
        for k in set_inactive_keys:
            self.selected_keys.remove(k)

        if set_active_keys:
            data_list.append(
                StatusChangedContainer(set_active_keys, True))
        if set_inactive_keys:
            data_list.append(
                StatusChangedContainer(set_inactive_keys, False))
        return data_list

    def emit_upward(self, s: SignalContainer):
        self.emit_downward(s)

    def add_segment(self, segment: RoiParameters) -> SegmentSignalData:
        r_angle, r_angle_std = self.image.ring_angle, self.image.ring_angle_str
        if segment.key is not None:
            new_key = segment.key
            segment = segment._replace(angle=r_angle,
                                       angle_std=r_angle_std)
        else:
            new_key = self._get_new_key()
            segment = segment._replace(key=new_key, angle=r_angle,
                                       angle_std=r_angle_std)
        self.segments_dict[new_key] = segment
        return SegmentSignalData(segment)

    def _get_new_key(self):
        if self.segments_dict:
            return max(self.segments_dict.keys()) + 1
        else:
            return 0


class AppNode(object):
    @property
    def image(self):
        return self.signal_connector.image

    def __init__(self, signal_connector: 'SignalConnector'):
        self.signal_connector = signal_connector

    def get_lower_connector(self, name: str = None):
        return self.signal_connector.get_lower_connector(name)

    def set_beam_center(self, beam_center: tuple):
        self.image.set_beam_center(beam_center)
        self.signal_connector.emit_upward(SignalContainer().geometry_changed(0))

    def set_image(self, image: np.ndarray):
        self.image.set_image(image)
        sc = SignalContainer(app_node=self)
        sc.image_changed(0)
        sc.geometry_changed(0)

    def add_transformation(self, name: str):
        self.image.add_transformation(name)
        sc = SignalContainer()
        sc.transformation_added(0)
        sc.geometry_changed(0)
        self.signal_connector.emit_upward(sc)

    def set_image_limits(self, limits=None):
        if self.image.intensity_limits != limits:
            self.image.set_image_limits(limits)
            SignalContainer().intensity_limits_changed(0).send(self)



from PyQt5.QtWidgets import QMainWindow

from .dock_area import AppDockArea
from ..utils import center_widget, Icon


class GiwaxsProgram(QMainWindow):
    _MinimumSize = (500, 500)
    _InitSize = (1000, 700)

    def __init__(self):
        super(GiwaxsProgram, self).__init__()
        self.filepath = None
        self.__init_toolbar__()
        self.main_widget = AppDockArea()

        self.setCentralWidget(self.main_widget)
        self.setWindowTitle('GIWAXS analysis')
        self.setWindowIcon(Icon('window_icon'))
        self.setGeometry(0, 0, *self._InitSize)
        self.setMinimumSize(*self._MinimumSize)

        center_widget(self)
        self.show()

    def __init_toolbar__(self):
        # self.toolbar = self.addToolBar('File manager')
        #
        # open_image_action = QAction(Icon('add'), 'Open image', self)
        # open_image_action.setShortcut('Ctrl+A')
        # open_image_action.triggered.connect(self._open_image_dialog)
        # self.toolbar.addAction(open_image_action)

        docks_toolbar = self.addToolBar('Docks')

        control_widget = docks_toolbar.addAction(Icon('folder'), 'File widget')
        control_widget.triggered.connect(lambda: self.main_widget.show_hide_docks('file_widget'))

        control_widget = docks_toolbar.addAction(Icon('control_widget'), 'Control widget')
        control_widget.triggered.connect(lambda: self.main_widget.show_hide_docks('control'))

        radial_profile = docks_toolbar.addAction(Icon('radial_profile'), 'Radial profile')
        radial_profile.triggered.connect(lambda: self.main_widget.show_hide_docks('radial_profile'))

        angular_profile = docks_toolbar.addAction(Icon('angular_profile'), 'Angular profile')
        angular_profile.triggered.connect(lambda: self.main_widget.show_hide_docks('angular_profile'))

        interpolation = docks_toolbar.addAction(Icon('interpolate'), 'Polar interpolation')
        interpolation.triggered.connect(lambda: self.main_widget.show_hide_docks('interpolation'))





import logging
import json
from pathlib import Path

logger = logging.getLogger(__name__)

_CONFIG_FOLDER = Path(__file__).parents[1] / 'user_config'
_DEFAULT_CONFIG_FOLDER = Path(__file__).parents[1] / 'static' / 'default_config'

if not _CONFIG_FOLDER.is_dir():
    _CONFIG_FOLDER.mkdir()


def read_config(config_name: str, get_default_values: bool = False):
    filename = f'{config_name}.json'
    config_path = _CONFIG_FOLDER / filename
    config_default_path = _DEFAULT_CONFIG_FOLDER / filename
    if get_default_values:
        return _read_config(config_default_path)
    else:
        saved_dict = _read_config(config_path)
        if not saved_dict:
            return _read_config(config_default_path)
        else:
            return saved_dict


def _read_config(path: Path):
    try:
        with open(str(path), 'r') as fp:
            saved_dict = json.load(fp)
            if type(saved_dict) is not dict:
                logger.error(f'JSON object is not dict:\n{saved_dict}')
            else:
                return saved_dict
    except FileNotFoundError:
        return
    except Exception as err:
        logger.exception(err)
        return


def save_config(config_name: str, config_dict: dict):
    config_path = _CONFIG_FOLDER / f'{config_name}.json'
    try:
        with open(config_path, 'w') as fp:
            json.dump(config_dict, fp)
    except Exception as err:
        logger.exception(err)



# -*- coding: utf-8 -*-

import logging

from pathlib import Path

import h5py

from PyQt5.QtWidgets import (QTreeView, QFileDialog, QWidget,
                             QHBoxLayout, QLabel, QMenu)
from PyQt5.QtCore import Qt, QItemSelectionModel
from PyQt5.QtGui import QStandardItem, QStandardItemModel

from .utils import read_h5_dict
from .items import (file_item_factory, FolderGroupItem, RoiItem,
                    H5GiwaxsItem, H5FileItem, AbstractGroupItem,
                    AbstractFileItem, H5DatasetItem)

from ..basic_widgets import RoundedPushButton
from ..roi.roi_widgets import EmptyROI, FileWidgetRoi
from ..roi.roi_containers import BasicROIContainer
from ..signal_connection import SignalConnector, SignalContainer, StatusChangedContainer

from ...utils import Icon, RoiParameters, save_execute

logger = logging.getLogger(__name__)


class FileModel(QStandardItemModel):
    def __init__(self):
        super(FileModel, self).__init__()
        self.setHorizontalHeaderLabels([''])
        self.setRowCount(0)

    def add_folder(self, folder_path: Path):
        folder_item = FolderGroupItem(folder_path)
        self.appendRow(folder_item)
        return folder_item

    def add_file(self, filepath: Path, row: int = None):
        item = file_item_factory(filepath)
        if item:
            if row is None:
                self.appendRow(item)
            else:
                self.insertRow(row, item)

        return item


class FileWidget(BasicROIContainer, QTreeView):
    def __init__(self, signal_connector: SignalConnector, parent=None):
        BasicROIContainer.__init__(self, signal_connector)
        QTreeView.__init__(self, parent=parent)
        self._model = FileModel()
        self.setEditTriggers(QTreeView.NoEditTriggers)
        self.setModel(self._model)
        self.selectionModel().currentChanged.connect(self._on_clicked)
        # self.clicked.connect(self._on_clicked)
        self.current_dataset = None
        self._future_dataset = None
        self.customContextMenuRequested.connect(
            self.context_menu
        )
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.__init_ui__()
        self.show()

    def __init_ui__(self):

        add_file_button = RoundedPushButton(icon=Icon('data'), radius=30,
                                            background_color='transparent')
        add_file_button.clicked.connect(self._open_add_file_menu)
        add_folder_button = RoundedPushButton(icon=Icon('folder'), radius=30,
                                              background_color='transparent')
        add_folder_button.clicked.connect(self._open_add_folder_menu)
        layout = self._get_header_layout(QStandardItem(), 'Files')
        layout.addWidget(add_file_button)
        layout.addWidget(add_folder_button)

    @save_execute('File widget process signal failed.')
    def process_signal(self, s: SignalContainer):
        super().process_signal(s)
        if self.current_dataset:
            for _ in s.geometry_changed():
                self.current_dataset.properties_item.update(
                    beam_center=self.image.beam_center
                )
            for _ in s.intensity_limits_changed():
                self.current_dataset.properties_item.update(
                    intensity_limits=self.image.intensity_limits
                )
            for _ in s.transformation_added():
                self.current_dataset.properties_item.update(
                    transformations=self.image.transformation.transformation_list
                )
            for _ in s.scale_changed():
                self.current_dataset.properties_item.update(
                    scale=self.image.scale
                )

    def add_roi(self, params: RoiParameters):
        roi = self._get_roi(params)
        if roi:
            roi.value_changed.connect(
                lambda value: self.signal_connector.emit_upward(
                    SignalContainer().segment_moved(value)))
            roi.status_changed.connect(self.emit_status_changed)
            roi.arbitrary_signal.connect(self.signal_connector.emit_upward)
            if not self._future_dataset:
                self.roi_dict[params.key] = roi
            self._add_item(roi)
            return roi

    def _get_roi(self, params: RoiParameters) -> 'AbstractROI':
        if self.current_dataset or self._future_dataset:
            return FileWidgetRoi(params)
        else:
            return EmptyROI(params)

    def _add_item(self, roi: FileWidgetRoi or EmptyROI):
        if ((self.current_dataset or self._future_dataset) and
                isinstance(roi, FileWidgetRoi)):
            item = RoiItem(roi.value.name)
            roi.set_item(item)
            parent = self._future_dataset or self.current_dataset
            parent.appendRow(item)
            # self.setExpanded(parent.index(), True)
            return roi

    def _remove_item(self, roi: FileWidgetRoi or EmptyROI):
        if self.current_dataset and isinstance(roi, FileWidgetRoi):
            self.current_dataset.removeRow(roi.item.row())

    def _on_status_changed(self, sig: StatusChangedContainer):
        if self.current_dataset:
            if not sig.status:
                for k in sig.keys:
                    self.roi_dict[k].set_inactive()
                    self.selectionModel().select(
                        self.roi_dict[k].item.index(),
                        QItemSelectionModel.Deselect)
            else:
                for k in sig.keys:
                    self.roi_dict[k].set_active()
                    self.selectionModel().select(
                        self.roi_dict[k].item.index(),
                        QItemSelectionModel.Select)

    def _open_add_file_menu(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        filepath, _ = QFileDialog.getOpenFileName(
            self, 'Open image', '',
            'edf, tiff, h5, hdf5 files (*.tiff *.edf *.h5 *.hdf5)', options=options)
        if filepath:
            self._model.add_file(Path(filepath))

    def _open_add_folder_menu(self):
        options = QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks
        folder_path = QFileDialog.getExistingDirectory(
            self, 'Choose directory containing edf, tiff or h5 files', '',
            options=options)
        if folder_path:
            self._model.add_folder(Path(folder_path))

    def _get_header_layout(self, item: QStandardItem, label: str):
        header_widget = QWidget(self)
        layout = QHBoxLayout()
        header_widget.setLayout(layout)
        label_widget = QLabel(label)
        layout.addWidget(label_widget, alignment=Qt.AlignLeft)
        layout.addStretch(1)
        self._model.appendRow(item)
        self.setIndexWidget(item.index(), header_widget)
        return layout

    def _on_clicked(self, index):
        item = self._model.itemFromIndex(index)
        if isinstance(item, AbstractGroupItem) and not item.content_uploaded:
            item.update_content()
            self.setExpanded(item.index(), True)
        elif isinstance(item, AbstractFileItem):
            if item.should_parse_file:
                self._parse_h5_item(item)
            else:
                data = item.get_data()
                if self.current_dataset != item and data.ndim == 2:
                    self._change_image_item(item, data)
        elif (isinstance(item, RoiItem) and
              item.roi and item.parent() is self.current_dataset):
            item.roi.send_active()

    @save_execute('Could not read saved h5 image.', silent=False)
    def _parse_h5_item(self, item: H5GiwaxsItem):
        try:
            self._future_dataset = item
            key = item.h5_key
            with h5py.File(item.filepath, 'r') as f:
                group = f[key]
                if 'image' not in group.keys():
                    return
                data = group['image'][()]
                if data.ndim != 2:
                    return
                item.properties_item.update(**read_h5_dict(group.attrs))
                roi_key = 0
                for k in group.keys():
                    dset = group[k]
                    if 'radius' in dset.attrs.keys() and 'width' in dset.attrs.keys():
                        params_dict = read_h5_dict(dset.attrs)
                        params = RoiParameters(**params_dict, key=roi_key)
                        self.add_roi(params)
                        roi_key += 1
        finally:
            self._future_dataset = None
            item.should_parse_file = False
        self._change_image_item(item, data)

    def _change_image_item(self, item, data):
        self.image.set_image(data)
        sc_delete = SignalContainer(app_node=self)
        sc_create = SignalContainer(app_node=self)
        sc_create.geometry_changed(0)
        sc_create.image_changed(0)
        for roi in self.roi_dict.values():
            roi.set_inactive()
            sc_delete.segment_deleted(roi.value, signal_type='exceptForNames')
        self.roi_dict = dict()
        self.current_dataset = item
        for child_item in self.current_dataset.get_child_rois():
            value = child_item.roi.value
            self.roi_dict[value.key] = child_item.roi
            sc_create.segment_created(value, signal_type='exceptForNames')
        if self.current_dataset.has_properties:
            self._set_file_properties_to_image()
        else:
            self._set_init_properties_to_file()

        sc_delete.send()
        sc_create.send()

    def _set_file_properties_to_image(self):
        properties = self.current_dataset.properties_item.get_dict()
        if 'intensity_limits' in properties.keys():
            self.image.set_image_limits(properties['intensity_limits'])
        if 'beam_center' in properties.keys():
            self.image.set_beam_center(properties['beam_center'])
        if 'transformations' in properties.keys():
            for name in properties['transformations']:
                self.image.add_transformation(name)
        if 'scale' in properties.keys():
            self.image.set_scale(properties['scale'])

    def _set_init_properties_to_file(self):
        self.current_dataset.properties_item.update(
            beam_center=self.image.beam_center,
            transformations=self.image.transformation.transformation_list,
            intensity_limits=self.image.intensity_limits,
            scale=self.image.scale
        )

    def context_menu(self, position):
        item = self._model.itemFromIndex(self.indexAt(position))
        menu = QMenu()
        if isinstance(item, FolderGroupItem):
            update_folder = menu.addAction('Update folder')
            update_folder.triggered.connect(lambda: self.update_group(item))
            close_folder = menu.addAction('Close folder')
            close_folder.triggered.connect(lambda: self._on_closing_group(item))
        elif isinstance(item, H5FileItem):
            update_folder = menu.addAction('Update h5 file')
            update_folder.triggered.connect(lambda: self.update_group(item))
            close_folder = menu.addAction('Close h5 file')
            close_folder.triggered.connect(lambda: self._on_closing_group(item))
        elif isinstance(item, AbstractFileItem):
            save_menu = menu.addMenu('Save')
            save_as_h5 = save_menu.addAction('Save as h5 file')
            save_to_h5 = save_menu.addAction('Save to existing h5 file')
            save_as_h5.triggered.connect(item.save_as_h5)
            save_to_h5.triggered.connect(item.save_to_h5)
            if isinstance(item, H5DatasetItem):
                save_here = save_menu.addAction('Save to current h5 file')
                save_here.triggered.connect(item.save_here)
        else:
            return
        menu.exec_(self.viewport().mapToGlobal(position))

    def _on_closing_group(self, item: H5FileItem or FolderGroupItem):
        if self._group_contains_current_dataset(item):
            self.current_dataset = None
            for k, v in self.roi_dict.items():
                self.roi_dict[k] = EmptyROI(v.value)
        item.close()

    def _group_contains_current_dataset(self, item: H5FileItem or FolderGroupItem):
        return (
                self.current_dataset and
                (isinstance(item, H5FileItem) and
                 item.filepath == self.current_dataset.filepath
                 or
                 isinstance(item, FolderGroupItem) and
                 item.filepath in self.current_dataset.filepath.parents)
        )

    @save_execute('Error occured while trying to update folder.', silent=False)
    def update_group(self, item: H5FileItem or FolderGroupItem):
        if self._group_contains_current_dataset(item):
            self.current_dataset = None
            for k, v in self.roi_dict.items():
                self.roi_dict[k] = EmptyROI(v.value)
        item.removeRows(0, item.rowCount())
        item.update_content()
        self.setExpanded(item.index(), True)



import logging
from functools import wraps
from typing import NamedTuple
from pathlib import Path
from enum import Enum

from PyQt5.QtWidgets import (QGraphicsColorizeEffect, QLineEdit,
                             QWidget, QApplication, QMessageBox)
from PyQt5.QtCore import QPropertyAnimation, Qt
from PyQt5.QtGui import QColor, QIcon

ICON_PATH = Path(__file__).parents[0] / 'static' / 'icons'

logger = logging.getLogger(__name__)


class RoiTypes(Enum):
    ring = 1
    segment = 2


class RoiParameters(NamedTuple):
    radius: float
    width: float
    angle: float = 180
    angle_std: float = 360
    orientations: list = None
    key: int = None
    name: str = None
    movable: bool = True
    fitted: bool = False
    fit_r_parameters: tuple = None
    type: str = RoiTypes.ring

    roi_types = RoiTypes  # not a field!


def save_execute(message: str = '', *, errors: tuple = None,
                 silent: bool = True, error_title: str = 'Error'):
    if not errors:
        errors = Exception

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except errors as err:
                logger.exception(f'Error while save_execute of {func.__name__}'
                                 f' with arguments {args}, {kwargs}:\n'
                                 f'{message}:\n{err}')
                if not silent:
                    show_error(message, error_title)

        return wrapper

    return decorator


def show_error(err: str, error_title: str):
    mb = QMessageBox()
    mb.setWindowTitle(error_title)
    logger.info(f'Error message shown: {err}.')
    mb.setWindowIcon(Icon('error'))
    mb.setText(err)
    mb.exec_()


class Icon(QIcon):
    def __init__(self, name: str):
        if name.find('.') == -1:
            name += '.png'
        name = str(ICON_PATH / name)
        QIcon.__init__(self, name)


def center_widget(widget):
    frame_gm = widget.frameGeometry()
    screen = QApplication.desktop().screenNumber(
        QApplication.desktop().cursor().pos())
    center_point = QApplication.desktop().screenGeometry(
        screen).center()
    frame_gm.moveCenter(center_point)
    widget.move(frame_gm.topLeft())


def validate_scientific_value(q_line_edit: QLineEdit,
                              data_type: type or None = float,
                              empty_possible: bool = False,
                              additional_conditions: tuple = ()):
    text_value = q_line_edit.text().replace(',', '.')
    if data_type is None:
        return text_value
    try:
        value = data_type(text_value)
    except ValueError:
        if not empty_possible:
            color_animation(q_line_edit)
        return
    for condition in additional_conditions:
        if not condition(value):
            return
    return value


def color_animation(widget: QWidget, color=Qt.red):
    effect = QGraphicsColorizeEffect(widget)
    widget.setGraphicsEffect(effect)

    widget.animation = QPropertyAnimation(effect, b"color")

    widget.animation.setStartValue(QColor(color))
    widget.animation.setEndValue(QColor(Qt.black))

    widget.animation.setLoopCount(1)
    widget.animation.setDuration(1500)
    widget.animation.start()



import sys
import logging
from PyQt5.QtWidgets import QApplication

from .gui import GiwaxsProgram

__all__ = ['GiwaxsProgram', 'run']


def run():
    # TODO: add logging config
    logging.basicConfig(level=logging.ERROR)
    app = QApplication(sys.argv)
    window = GiwaxsProgram()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
import logging
from abc import abstractmethod
from copy import deepcopy
from pathlib import Path

import numpy as np
import h5py

from PyQt5.QtGui import QStandardItem

from .utils import (save_as_h5_dialog, save_to_h5_dialog,
                    save_create_h5_subgroup,
                    prepare_dict_to_h5, parse_h5_group,
                    filter_dirs, filter_files)
from ...read_data import get_image_from_path
from ...utils import Icon, save_execute

logger = logging.getLogger(__name__)

_H5_GIWAXS_DATA_KEY = '__GIWAXS_DATA__'


class AbstractItem(QStandardItem):
    def __init__(self, filepath: Path, *args, **kwargs):
        self.filepath = filepath
        super().__init__(self.__get_name__())

    def __get_name__(self):
        return self.filepath.name


class AbstractFileItem(AbstractItem):
    should_parse_file = False

    @property
    def properties_item(self):
        if not self._properties_item:
            self._properties_item = PropertiesItem('Properties')
            self.insertRow(0, self.properties_item)
        return self._properties_item

    @property
    def has_properties(self):
        return self._properties_item is not None

    def __init__(self, filepath: Path, *args, **kwargs):
        super().__init__(filepath)
        self._properties = dict()
        self._properties_item = None
        self.setIcon(Icon('data'))

    def get_child_rois(self):
        for row in range(self.rowCount()):
            item = self.child(row)
            if isinstance(item, RoiItem) and item.roi:
                yield item

    @abstractmethod
    def get_data(self) -> np.array:
        pass

    @save_execute('An error occured while trying to save file.',
                  silent=False, error_title='Save file error')
    def save_as_h5(self, *args):
        filepath = save_as_h5_dialog()
        if filepath:
            with h5py.File(filepath, 'w') as f:
                self._save_to_h5(f)

    @save_execute('An error occured while trying to save file.',
                  silent=False, error_title='Save file error')
    def save_to_h5(self, *args):
        filepath = save_to_h5_dialog()
        if filepath:
            with h5py.File(filepath, 'a') as f:
                self._save_to_h5(f)

    def _save_to_h5(self, f: h5py.File, data: np.ndarray = None, name: str = None):
        if data is None:
            data = self.get_data()
        # TODO: fix problem with saving to h5!
        name = name or self.__get_name__().split('.')[0]
        group = save_create_h5_subgroup(f, name)
        group.attrs.update({_H5_GIWAXS_DATA_KEY: 1})
        group.attrs.update(prepare_dict_to_h5(self.properties_item.get_dict()))
        group.create_dataset('image', data=data)
        for i, item in enumerate(self.get_child_rois()):
            d = prepare_dict_to_h5(item.get_dict())
            dset = save_create_h5_subgroup(group, d.get('name', 'segment'),
                                           data=d.pop('key', 0))
            dset.attrs.update(d)


class AbstractGroupItem(AbstractItem):
    def __init__(self, filepath: Path, *args, **kwargs):
        super(AbstractGroupItem, self).__init__(filepath)
        self.setIcon(Icon('folder'))
        self.content_uploaded = False

    def update_content(self):
        self.content_uploaded = True
        return self._update_content()

    @abstractmethod
    def _update_content(self):
        pass


class AttributesItem(QStandardItem):
    @abstractmethod
    def get_dict(self):
        pass


class RoiItem(AttributesItem):
    def __init__(self, name: str):
        self.roi = None
        QStandardItem.__init__(self, name)
        self.setIcon(Icon('roi_item'))

    def get_dict(self):
        if not self.roi:
            return dict()
        else:
            return self.roi.value._asdict()


class PropertiesItem(AttributesItem):
    def __init__(self, *args):
        super().__init__(*args)
        self._properties = dict()
        self.setIcon(Icon('properties'))

    def update(self, **kwargs):
        self._properties.update(kwargs)

    def get_dict(self):
        return deepcopy(self._properties)


class H5Item(AbstractItem):
    def __init__(self, filepath: Path, h5_key: str, *args, **kwargs):
        self.h5_key = h5_key
        super(H5Item, self).__init__(filepath)

    def __get_name__(self):
        return self.h5_key.split('/')[-1]


class H5GroupItem(H5Item, AbstractGroupItem):
    def __init__(self, filepath: Path, h5_key: str, *args, **kwargs):
        kwargs['filepath'] = filepath
        kwargs['h5_key'] = h5_key
        super(H5GroupItem, self).__init__(*args, **kwargs)
        self.setIcon(Icon('h5_group_folder'))
        # H5Item.__init__(self, filepath, h5_key)

    @save_execute('Could not read h5 file.', silent=False)
    def _update_content(self):
        with h5py.File(self.filepath, 'r') as f:
            for h5item in parse_h5_group(f, self.h5_key):
                item = h5_item_factory(h5item, self.filepath)
                if item:
                    self.appendRow(item)


class H5FileItem(H5GroupItem):
    def __init__(self, filepath: Path, *args, **kwargs):
        kwargs['filepath'] = filepath
        kwargs['h5_key'] = ''
        super(H5FileItem, self).__init__(*args, **kwargs)
        self.setIcon(Icon('h5_folder'))

    def __get_name__(self):
        return self.filepath.name

    def close(self):
        parent = self.parent() or self.model()
        parent.removeRow(self.row())


class H5DatasetItem(H5Item, AbstractFileItem):
    @save_execute('Error while trying to get data from h5 file.', silent=True)
    def get_data(self):
        with h5py.File(self.filepath) as f:
            return f[self.h5_key][()]

    def save_here(self):
        with h5py.File(self.filepath) as f:
            data = self.get_data()
            name = self.text()
            del f[self.h5_key]
            # with h5py.File(self.filepath) as f:
            self._save_to_h5(f, data, name)


class H5GiwaxsItem(H5DatasetItem):
    should_parse_file = True

    @save_execute('Error while trying to get data from h5 file.', silent=True)
    def get_data(self):
        with h5py.File(self.filepath, 'r') as f:
            return f[f'{self.h5_key}/image'][()]


class FolderGroupItem(AbstractGroupItem):
    def _update_content(self):
        for dirpath in filter_dirs(self.filepath):
            self.appendRow(FolderGroupItem(dirpath))
        for filepath in filter_files(self.filepath):
            item = file_item_factory(filepath)
            if item:
                self.appendRow(item)

    def close(self):
        parent = self.parent() or self.model()
        parent.removeRow(self.row())


class EdfFileItem(AbstractFileItem):
    @save_execute('Cound not read from edf file', silent=False)
    def get_data(self):
        return get_image_from_path(self.filepath)


class TiffFileItem(AbstractFileItem):
    @save_execute('Cound not read from tiff file', silent=False)
    def get_data(self):
        return get_image_from_path(self.filepath)


def file_item_factory(filepath: Path):
    if filepath.suffix == '.edf':
        return EdfFileItem(filepath)
    elif filepath.suffix in ['.tif', '.tiff']:
        return TiffFileItem(filepath)
    elif filepath.suffix == '.h5':
        return H5FileItem(filepath)


def h5_item_factory(h5item: h5py.Group or h5py.Dataset, filepath: Path):
    if isinstance(h5item, h5py.Group):
        if _H5_GIWAXS_DATA_KEY in h5item.attrs.keys():
            return H5GiwaxsItem(filepath, h5item.name)
        else:
            return H5GroupItem(filepath, h5item.name)
    elif isinstance(h5item, h5py.Dataset):
        return H5DatasetItem(filepath, h5item.name)



# -*- coding: utf-8 -*-

import logging
from pathlib import Path

import numpy as np
import h5py

from PyQt5.QtWidgets import QFileDialog

logger = logging.getLogger(__name__)

_AVAILABLE_FILE_FORMATS = tuple('.tif .tiff .h5 .hdf5 .edf'.split())


def filter_files(path: Path):
    yield from (p for p in path.iterdir() if p.suffix in _AVAILABLE_FILE_FORMATS)


def filter_dirs(path: Path):
    yield from (p for p in path.iterdir() if p.is_dir())


def parse_h5_group(file: h5py.File, key: str):
    group = file[key] if key else file
    yield from (group[k] for k in group.keys())


def save_as_h5_dialog():
    return QFileDialog.getSaveFileName(None, 'Create h5 file', '', 'hdf5 files (*.h5)')[0]


def save_to_h5_dialog():
    options = QFileDialog.Options()
    options |= QFileDialog.DontUseNativeDialog
    filename, _ = QFileDialog.getOpenFileName(
        None, 'Choose h5 file', '',
        'hdf5 files (*.h5)', options=options)
    return filename


def save_create_h5_subgroup(group: h5py.Group, name: str, data=None):
    init_name = name
    i = 1
    while name in group.keys():
        name = f'{init_name}_{i}'
        i += 1
    if data is not None:
        return group.create_dataset(name, data=data)
    else:
        return group.create_group(name)


def prepare_dict_to_h5(d: dict):
    d = dict(d)
    for k, v in d.items():
        if isinstance(v, str):
            d[k] = v.encode()
        elif v is None:
            d[k] = 'none'.encode()
        elif isinstance(v, list) or isinstance(v, tuple):
            d[k] = [item.encode() if isinstance(item, str) else item
                    for item in v if item is not None]
    return d


def read_h5_dict(d):
    d = dict(d)
    for k, v in d.items():
        if isinstance(v, bytes):
            if v == b'none':
                d[k] = None
            else:
                d[k] = v.decode()
        elif isinstance(v, np.ndarray):
            d[k] = [item.decode() if isinstance(item, bytes) else item
                    for item in v]
    return d



# -*- coding: utf-8 -*-
import logging

from PyQt5.QtWidgets import QMainWindow, QAction

from .parameters_widget import InterpolateSetupWindow

from ..basic_widgets import CustomImageViewer
from ..roi.roi_widgets import EmptyROI, Roi2DRect
from ..roi.roi_containers import AbstractROIContainer
from ..signal_connection import SignalConnector, SignalContainer

from ...utils import RoiParameters, Icon

logger = logging.getLogger(__name__)


# TODO: fix bug with scaling

# TODO: fix bug with negative angles

# TODO: add axes


class InterpolateImageWidget(AbstractROIContainer, QMainWindow):
    def __init__(self, signal_connector: SignalConnector,
                 parent=None):
        AbstractROIContainer.__init__(self, signal_connector)
        QMainWindow.__init__(self, parent)
        self._setup_window = None
        self._image_viewer = CustomImageViewer(self)
        self.setCentralWidget(self._image_viewer)
        self.__init_toolbar__()
        self.update_image()

    def process_signal(self, s: SignalContainer):
        super().process_signal(s)
        for _ in s.image_changed():
            self.update_image()
        for _ in s.geometry_changed_finish():
            self.update_image()

    def _add_item(self, roi):
        if isinstance(roi, Roi2DRect):
            self._image_viewer.image_plot.addItem(roi)

    def _remove_item(self, roi):
        if isinstance(roi, Roi2DRect):
            self._image_viewer.image_plot.removeItem(roi)

    def _get_roi(self, params: RoiParameters):
        r_size, phi_size = (self.image.interpolation.r_size,
                            self.image.interpolation.phi_size)
        image, rr, phi = self.image.image, self.image.rr, self.image.phi
        if any(x is None for x in (r_size, phi_size, image, rr, phi)):
            return EmptyROI(params)
        return Roi2DRect(params, rr, r_size, phi, phi_size)

    def __init_toolbar__(self):

        setup_toolbar = self.addToolBar('Setup')
        setup_toolbar.setStyleSheet('background-color: black;')

        setup_action = setup_toolbar.addAction(Icon('setup'), 'Setup')
        setup_action.triggered.connect(self.open_setup_window)

    def update_image(self):
        p_image = self.image.interpolate()
        if p_image is not None:
            roi_values = [value.parameters for value in self.roi_dict.values()]
            for p in roi_values:
                self.delete_roi(p)
            self.set_data(p_image)
            for p in roi_values:
                self.add_roi(p)

    def set_data(self, image):
        self._image_viewer.set_data(image)

    def open_setup_window(self):
        self._setup_window = InterpolateSetupWindow()
        self._setup_window.apply_signal.connect(self.set_parameters)
        self._setup_window.close_signal.connect(self.close_setup)
        self._setup_window.show()

    def set_parameters(self, params: dict):
        self.image.set_interpolation_parameters(params)
        self.update_image()

    def close_setup(self):
        self._setup_window = None



from typing import NamedTuple
import logging

import numpy as np

from .exceptions import UnknownTransformation
from .interpolation.interpolation import Interpolation

logger = logging.getLogger(__name__)


class Geometry(NamedTuple):
    xx: np.ndarray = None
    yy: np.ndarray = None
    rr: np.ndarray = None
    phi: np.ndarray = None
    beam_center: tuple = None

    @classmethod
    def get(cls, shape: tuple, center: tuple):
        xx, yy = np.meshgrid(
            np.arange(shape[1]) - center[1],
            np.arange(shape[0]) - center[0]
        )
        rr = np.sqrt(xx ** 2 + yy ** 2)
        phi = np.arctan2(yy, xx)
        return cls(xx=xx, yy=yy, rr=rr, phi=phi, beam_center=center)


class ImageScale(NamedTuple):
    scale: float = 1.
    unit: str = ''
    previous_scale: float = 1.


class RingAngles(NamedTuple):
    angle: float = None
    angle_std: float = None


class ImageTransformation(object):
    @property
    def transformation_list(self):
        return self._transformation_list

    def __init__(self):
        self._transformation_list = list()
        self._transformation_dict = dict(
            horizontal=self.horizontal,
            vertical=self.vertical,
            rotate_right=self.rotate_right,
            rotate_left=self.rotate_left)

    def add_transformation(self, name: str):
        if name not in self._transformation_dict.keys():
            raise UnknownTransformation()
        self._transformation_list.append(name)
        # TODO: clever search, delete opposite transformations

    def transform(self, image):
        for t in self._transformation_list:
            image = self._transformation_dict[t](image)
        return image

    def last_transform(self, image):
        if self._transformation_list:
            return self._transformation_dict[self._transformation_list[-1]](image)
        else:
            return image

    def clear(self):
        self._transformation_list = list()

    @staticmethod
    def horizontal(image):
        return np.flip(image, axis=1)

    @staticmethod
    def vertical(image):
        return np.flip(image, axis=0)

    @staticmethod
    def rotate_right(image):
        return np.rot90(image, k=-1)

    @staticmethod
    def rotate_left(image):
        return np.rot90(image, k=1)


class Image(object):
    @property
    def image(self):
        return self._image

    @property
    def shape(self):
        return self._image.shape if self._image is not None else None

    @property
    def geometry(self):
        return self._geometry

    @property
    def beam_center(self):
        return self._beam_center

    @property
    def rr(self):
        return self.geometry.rr

    @property
    def phi(self):
        return self.geometry.phi

    @property
    def xx(self):
        return self.geometry.xx

    @property
    def yy(self):
        return self.geometry.yy

    @property
    def intensity_limits(self):
        return self._intensity_limits

    @property
    def scale(self):
        return self._scale.scale

    @property
    def scale_unit(self):
        return self._scale.unit

    @property
    def scale_change(self):
        return self.scale / self._scale.previous_scale

    @property
    def ring_angle(self):
        return self._ring_angles.angle

    @property
    def ring_angle_str(self):
        return self._ring_angles.angle_std

    @property
    def interpolation(self):
        return self._interpolation

    def __init__(self):
        self._source_image = None
        self._image = None
        self.transformation = ImageTransformation()
        self._intensity_limits = None
        self._keep_limits = True
        self.save_transformation = False
        self._beam_center = (0, 0)
        self._geometry = Geometry()
        self._scale = ImageScale()
        self._ring_angles = RingAngles()
        self._interpolation = Interpolation()

    def set_image_limits(self, limits: tuple = None):
        self._intensity_limits = limits
        if not limits:
            self._keep_limits = False
        else:
            self._keep_limits = True

    def set_beam_center(self, beam_center: tuple):
        if beam_center:
            self._beam_center = beam_center
            self.update_geometry()

    def add_transformation(self, name):
        if self._image is None:
            return
        try:
            self.transformation.add_transformation(name)
        except UnknownTransformation:
            logger.exception('')
            return
        self._image = self.transformation.last_transform(self._image)
        self.update_geometry()

    def set_image(self, image):
        if not isinstance(image, np.ndarray) or image.ndim != 2:
            logger.error(f'Set image got wrong argument: {image}')
            return
        if not self.save_transformation:
            self.transformation.clear()
        if not self._keep_limits:
            self._intensity_limits = None
        self._source_image = image
        self._image = self.transformation.transform(image)
        self.update_geometry()

    def update_geometry(self):
        if self._image is None or self._beam_center is None:
            return
        self._geometry = Geometry.get(self.shape, self._beam_center)
        phi = self._geometry.phi
        self._ring_angles = RingAngles(
            angle=(phi.max() + phi.min()) / 2 * 180 / np.pi,
            angle_std=(phi.max() - phi.min()) * 180 / np.pi
        )
        self.interpolation.set_geometry(self.geometry)

    def set_scale(self, scale: float, unit: str = ''):
        self._scale = ImageScale(scale, unit, self.scale)

    def set_interpolation_parameters(self, parameters: dict):
        self.interpolation.set_parameters(parameters)

    def interpolate(self):
        return self.interpolation.interpolate(self.image)

    def get_angular_profile(self, r1: float, r2: float):
        # TODO: refactor! should be handled by Interpolation class
        image, rr, r_size, p = (self.image, self.rr, self.interpolation.r_size,
                                self.interpolation.interpolation_geometry.p)
        if any(x is None for x in (image, rr, r_size, p)):
            return
        r_min = rr.min()
        r_ratio = (rr.max() - rr.min()) / r_size
        r1 = int((r1 / self.scale - r_min) / r_ratio)
        r2 = int((r2 / self.scale - r_min) / r_ratio)
        x = p * 180 / np.pi
        return x, self.interpolation.get_angular_profile(r1, r2)



# -*- coding: utf-8 -*-
import logging

import numpy as np

from pyqtgraph import CircleROI, LineSegmentROI
from PyQt5.QtWidgets import (QMainWindow, QWidget, QVBoxLayout)
from PyQt5.QtCore import pyqtSignal, Qt

from .basic_widgets import CustomImageViewer, AnimatedSlider
from .signal_connection import SignalConnector, SignalContainer, AppNode
from .roi.roi_widgets import Roi2DRing
from .roi.roi_containers import AbstractROIContainer
from ..utils import Icon, center_widget, RoiParameters

logger = logging.getLogger(__name__)


class GiwaxsImageViewer(AbstractROIContainer, CustomImageViewer):
    @property
    def beam_center(self):
        return self.image.beam_center

    class BeamCenterRoi(CircleROI):
        _ROI_SIZE = 200

        def __init__(self, beam_center, parent):
            CircleROI.__init__(self, (beam_center[1], beam_center[0]),
                               self._ROI_SIZE, movable=False, parent=parent)
            self._center = None
            self.set_center(beam_center)

        def set_center(self, value: tuple, y=None, update=True, finish=True, ):
            self._center = value
            radius = self.size().x() / 2
            pos = (value[1] - radius, value[0] - radius)
            super(GiwaxsImageViewer.BeamCenterRoi, self).setPos(
                pos, y, update, finish)

        def set_size(self, size: float):
            self.setSize((size, size), update=False, finish=False)
            self.set_center(self._center)

        def set_scale(self, scale: float):
            size = self._ROI_SIZE * scale
            self.set_size(size)

    class ZeroAngleRoi(LineSegmentROI):
        _ROI_SIZE = 300

        def __init__(self, beam_center, zero_angle, invert_angle, parent):
            self.beam_center = beam_center
            self.zero_angle = 0
            self.invert_angle = invert_angle

            self.coords = self.beam_center, self._get_second_point_coordinates()
            LineSegmentROI.__init__(self, self.coords, parent=parent, movable=False)
            self.handles[0]['item'].mouseDragEvent = lambda *args: None
            self.handles[0]['item'].mouseClickEvent = lambda *args: None
            self.handles[1]['item'].mouseDragEvent = lambda *args: None
            self.handles[1]['item'].mouseClickEvent = lambda *args: None

            self.set_center(self.beam_center)
            self.set_angle(zero_angle)
            self.hide()

        def _get_second_point_coordinates(self):
            angle = self.zero_angle
            if self.invert_angle:
                angle *= -1
            x = self.beam_center[1] + np.cos(angle * np.pi / 180) * self._ROI_SIZE
            y = self.beam_center[0] + np.sin(angle * np.pi / 180) * self._ROI_SIZE
            return x, y

        def set_center(self, value):
            self.beam_center = value
            self.setPos((value[1], value[0]))

        def set_angle(self, value):
            if self.invert_angle:
                value *= -1
            self._set_angle(value)

        def _set_angle(self, angle):
            diff_angle, self.zero_angle = angle - self.zero_angle, angle
            self.rotate(diff_angle)

        def set_invert(self, value):
            if value != self.invert_angle:
                self.invert_angle = value
                self._set_angle(-self.zero_angle)

    def __init__(self, signal_connector: SignalConnector,
                 parent=None, **kwargs):
        AbstractROIContainer.__init__(self, signal_connector)
        CustomImageViewer.__init__(self, parent, **kwargs)

        self._geometry_params_widget = None
        self.hist.sigLevelChangeFinished.connect(self._on_limits_changed)
        self.__init_center_roi__()

    def process_signal(self, s: SignalContainer):
        AbstractROIContainer.process_signal(self, s)
        for _ in s.image_changed():
            self.set_data(self.image.image, change_limits=False)
            self.set_levels(self.image.intensity_limits)
        for _ in s.transformation_added():
            self.set_data(self.image.image, change_limits=False)
        for _ in s.geometry_changed():
            self.update_beam_center(self.image.beam_center, emit_value=False)

    def _on_limits_changed(self):
        levels = self.get_levels()
        if levels != (0, 1):
            self.set_image_limits(self.get_levels())

    def _on_scale_changed(self):
        scale = self.image.scale
        self.set_scale(scale)
        self.center_roi.set_scale(scale)

    def _get_roi(self, params: RoiParameters):
        return Roi2DRing(params)

    def _add_item(self, roi):
        self.image_plot.addItem(roi)

    def _remove_item(self, roi):
        self.image_plot.removeItem(roi)

    def __init_center_roi__(self):
        self.center_roi = self.BeamCenterRoi(self.beam_center, parent=self.image_item)
        self.center_roi.setZValue(10)
        self.image_plot.addItem(self.center_roi)

        self.angle_roi = self.ZeroAngleRoi(self.beam_center, 0, False, self.image_item)
        self.angle_roi.setZValue(10)
        self.image_plot.addItem(self.angle_roi)

    def open_geometry_parameters(self):

        logger.debug(self._geometry_params_widget)
        if self.image.image is not None and self._geometry_params_widget is None:
            self._geometry_params_widget = GeometryParametersWidget(
                self.image.shape, self.beam_center, scale=self.image.scale)
            self._geometry_params_widget.change_center.connect(
                self.update_beam_center)
            self._geometry_params_widget.change_zero_angle.connect(self.set_zero_angle)
            self._geometry_params_widget.change_invert_angle.connect(self.set_invert_angle)
            self._geometry_params_widget.scale_changed.connect(self.emit_scale_changed)
            self._geometry_params_widget.close_event.connect(self.on_closing_beam_center_widget)

    def on_closing_beam_center_widget(self):
        self._geometry_params_widget = None
        SignalContainer(app_node=self).geometry_changed_finish(0).send()

    def emit_scale_changed(self, value):
        self.image.set_scale(value)
        SignalContainer(app_node=self).scale_changed(0).send()

    def set_zero_angle(self, value):
        self.angle_roi.set_angle(value)

    def set_invert_angle(self, value):
        self.angle_roi.set_invert(value)

    def update_beam_center(self, value, emit_value: bool = True):
        if emit_value:
            self.set_beam_center(value)
        self.set_center((value[1], value[0]), pixel_units=True)


class GeometryParametersWidget(QWidget):
    change_center = pyqtSignal(list)
    change_zero_angle = pyqtSignal(float)
    change_invert_angle = pyqtSignal(bool)
    scale_changed = pyqtSignal(float)

    close_event = pyqtSignal()

    def __init__(self, image_shape: tuple,
                 beam_center: tuple, zero_angle: float = 0,
                 angle_direction: bool = True, scale: float = 1):
        super(GeometryParametersWidget, self).__init__(None, Qt.WindowStaysOnTopHint)
        self.beam_center = list(beam_center)
        self.image_shape = image_shape
        self.zero_angle = zero_angle
        self.scale = scale
        self.angle_direction = angle_direction
        self.__init__ui__()
        self.setWindowTitle('Set beam center coordinates')
        self.setWindowIcon(Icon('setup'))
        center_widget(self)
        self.show()

    def closeEvent(self, a0) -> None:
        self.close_event.emit()
        QWidget.closeEvent(self, a0)

    def __init__ui__(self):
        layout = QVBoxLayout(self)

        self.x_slider = AnimatedSlider('X center', (0, self.image_shape[1]),
                                       self.beam_center[1], self,
                                       Qt.Horizontal, disable_changing_status=True)
        self.x_slider.valueChanged.connect(self._connect_func(1))

        self.y_slider = AnimatedSlider('Y center', (0, self.image_shape[0]),
                                       self.beam_center[0], self,
                                       Qt.Horizontal, disable_changing_status=True)
        self.y_slider.valueChanged.connect(self._connect_func(0))

        # self.angle_slider = AnimatedSlider('Zero angle', (0, 360),
        #                                    self.zero_angle, self,
        #                                    Qt.Horizontal, disable_changing_status=True)
        # self.angle_slider.valueChanged.connect(self._connect_func(2))
        #
        # self.invert_angle_box = QCheckBox('Invert angle')
        # self.invert_angle_box.toggled.connect(self._connect_func(3))

        self.scale_edit = AnimatedSlider('Q to pixel ratio', (1e-10, 10),
                                         self.scale, self,
                                         Qt.Horizontal, disable_changing_status=True,
                                         decimals=5)
        self.scale_edit.valueChanged.connect(self.on_scale_changed)

        layout.addWidget(self.x_slider)
        layout.addWidget(self.y_slider)
        layout.addWidget(self.scale_edit)
        # layout.addWidget(self.angle_slider)
        # layout.addWidget(self.invert_angle_box)

    def on_scale_changed(self, value):
        self.scale = value
        self.scale_changed.emit(value)

    def _connect_func(self, ind: int):
        def beam_center_changed(value):
            self.beam_center[ind] = value
            self.change_center.emit(self.beam_center)

        def angle_zero_changed(value):
            self.zero_angle = value
            self.change_zero_angle.emit(self.zero_angle)

        def angle_direction_changed(value):
            self.angle_direction = value
            self.change_invert_angle.emit(self.angle_direction)

        if ind < 2:
            return beam_center_changed
        elif ind == 2:
            return angle_zero_changed
        else:
            return angle_direction_changed


class Basic2DImageWidget(AppNode, QMainWindow):

    def __init__(self, signal_connector, parent=None):
        AppNode.__init__(self, signal_connector)
        QMainWindow.__init__(self, parent)
        self.image_viewer = GiwaxsImageViewer(self.get_lower_connector(), self)
        self.setCentralWidget(self.image_viewer)
        self.__init_toolbar__()

    def __init_toolbar__(self):
        toolbar = self.addToolBar('Main')
        toolbar.setStyleSheet('background-color: black;')

        rotate_action = toolbar.addAction(Icon('rotate'), 'Rotate')
        rotate_action.triggered.connect(
            lambda: self.add_transformation('rotate_right'))

        flip_h = toolbar.addAction(Icon('flip_horizontal'), 'Horizontal flip')
        flip_h.triggered.connect(
            lambda: self.add_transformation('horizontal'))

        flip_v = toolbar.addAction(Icon('flip_vertical'), 'Vertical flip')
        flip_v.triggered.connect(
            lambda: self.add_transformation('vertical'))

        geometry_toolbar = self.addToolBar('Geometry')
        geometry_toolbar.setStyleSheet('background-color: black;')

        set_beam_center_action = toolbar.addAction(
            Icon('center'), 'Beam center')
        set_beam_center_action.triggered.connect(
            self.centralWidget().open_geometry_parameters)



# -*- coding: utf-8 -*-


class UnknownTransformation(ValueError):
    pass


class KeySignalNameError(ValueError):
    pass


class SignalNotFoundError(ValueError):
    pass


class AppNodeNotProvidedError(ValueError):
    pass



# -*- coding: utf-8 -*-
import weakref

import numpy as np

from PyQt5.QtCore import QObject, Qt, QRectF, pyqtSignal, QItemSelectionModel
from PyQt5.QtGui import QPainter, QPen, QPainterPath, QStandardItem, QColor
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QLineEdit, QGridLayout

from pyqtgraph import LinearRegionItem, RectROI, ROI

from .abstract_roi_widget import AbstractROI
from .roi_menu import RadialProfileContextMenu, RoiContextMenu
from ..basic_widgets import ControlSlider, RoundedPushButton, DeleteButton
from ...utils import RoiParameters, Icon


class EmptyROI(AbstractROI, QObject):
    def __init__(self, value: RoiParameters, *args, **kwargs):
        AbstractROI.__init__(self, value)
        QObject.__init__(self, *args, **kwargs)
        self.init_roi()

    @property
    def value(self):
        return self.parameters

    @value.setter
    def value(self, value: RoiParameters):
        self.parameters = value

    def set_active(self):
        pass

    def set_inactive(self):
        pass

    def set_fixed(self):
        pass

    def set_unfixed(self):
        pass

    def set_color(self, color):
        pass

    def set_movable(self, movable: bool):
        pass


class Roi1D(AbstractROI, LinearRegionItem):
    _ACTIVE_Z = -15
    _INACTIVE_Z = -25
    _FIXED_Z = -30

    def __init__(self, value: RoiParameters, *args, **kwargs):
        AbstractROI.__init__(self, value)
        LinearRegionItem.__init__(self, *args, **kwargs)
        self.sigRegionChanged.connect(self.roi_is_moving)
        self.init_roi()

    def roi_is_moving(self):
        if self.moving:
            self.send_value()
        for l in self.lines:
            if l.moving:
                self.send_value()

    @property
    def value(self):
        x1, x2 = self.getRegion()
        r, w = (x1 + x2) / 2, x2 - x1
        self.parameters = self.parameters._replace(radius=r, width=w)
        return self.parameters

    @value.setter
    def value(self, value: RoiParameters):
        if value != self.value:
            self.parameters = value
            r, w = value.radius, value.width
            x1, x2 = r - w / 2, r + w / 2
            self.setRegion((x1, x2))

    def set_fixed(self):
        self.setZValue(self._FIXED_Z)
        super().set_fixed()

    def set_color(self, color):
        self.setBrush(color)
        self.viewRangeChanged()

    def set_movable(self, movable: bool):
        self.setMovable(movable)

    def mouseDragEvent(self, ev):
        if not ev.modifiers() == Qt.ShiftModifier:
            LinearRegionItem.mouseDragEvent(self, ev)
        else:
            self.mouseClickEvent(ev)

    def mouseClickEvent(self, ev):
        if self.moving and ev.button() == Qt.RightButton:
            ev.accept()
            for i, l in enumerate(self.lines):
                l.setPos(self.startPositions[i])
            self.moving = False
            self.sigRegionChanged.emit(self)
            self.sigRegionChangeFinished.emit(self)
            self.send_value()
        elif ev.button() == Qt.RightButton:
            ev.accept()
            self.show_context_menu(ev)
        elif ev.button() == Qt.LeftButton and ev.modifiers() == Qt.ShiftModifier:
            ev.accept()
            self.change_active(False)
        elif ev.button() == Qt.LeftButton:
            self.change_active()
        self.viewRangeChanged()

    def show_context_menu(self, ev):
        RadialProfileContextMenu(self)

    def set_active(self):
        self.setZValue(self._ACTIVE_Z)
        super().set_active()

    def set_inactive(self):
        self.setZValue(self._INACTIVE_Z)
        super().set_inactive()


class Roi1DAngular(Roi1D):
    @property
    def value(self):
        x1, x2 = self.getRegion()
        a, a_w = (x1 + x2) / 2, x2 - x1
        self.parameters = self.parameters._replace(angle=a, angle_std=a_w)
        return self.parameters

    @value.setter
    def value(self, value: RoiParameters):
        if value != self.value:
            self.parameters = value
            a, a_w = value.angle, value.angle_std
            x1, x2 = a - a_w / 2, a + a_w / 2
            self.setRegion((x1, x2))


class Roi2DRect(AbstractROI, RectROI):
    _USE_BRIGHT_COLOR = True
    _g = 180 / np.pi
    # TODO: fix bug with active status
    # TODO: overwrite Handle class to manage moving and be able to fix/unfix it
    # TODO: fix moving in each direction separately

    def __init__(self, value: RoiParameters, rr, r_size, phi, phi_size):
        AbstractROI.__init__(self, value)
        RectROI.__init__(self, pos=(0, 0), size=(1, 1))
        self.__init_ratios__(rr, r_size, phi, phi_size)

        self.init_roi()
        self.handle = self.handles[0]['item']
        self.sigRegionChanged.connect(self.handle_is_moving)

    def __init_ratios__(self, rr, r_size, phi, phi_size):
        # TODO: refactor!! rois should not handle these problems
        # TODO: fix bug: not sensitive to scale changing
        _g = self._g
        self.r_ratio = (rr.max() - rr.min()) / r_size
        self.phi_ratio = (phi.max() - phi.min()) / phi_size * _g
        self.r_min = rr.min()
        self.phi_min = phi.min() * _g

    def handle_is_moving(self):
        if self.handle.isMoving:
            self.send_value()

    @property
    def value(self):
        size, pos = self.size(), self.pos()
        w, a_w = size
        r, a = pos[0] + w / 2, pos[1] + a_w / 2
        r *= self.r_ratio + self.r_min
        w *= self.r_ratio + self.r_min
        a *= self.phi_ratio + self.phi_min
        a_w *= self.phi_ratio + self.phi_min
        self.parameters = self.parameters._replace(
            radius=r, width=w, angle=a, angle_std=abs(a_w))
        return self.parameters

    @value.setter
    def value(self, value: RoiParameters):
        if value != self.parameters:
            self.parameters = value
            r, w = (value.radius - self.r_min) / self.r_ratio, (value.width - self.r_min) / self.r_ratio
            a = (value.angle - self.phi_min) / self.phi_ratio
            a_w = abs((value.angle_std - self.phi_min) / self.phi_ratio)

            pos = [r - w / 2, a - a_w / 2]
            size = [w, a_w]
            self.setSize(size)
            self.setPos(pos)

    def set_color(self, color):
        self.setPen(color)

    def set_movable(self, movable: bool):
        self.translatable = movable

    def mouseDragEvent(self, ev):
        RectROI.mouseDragEvent(self, ev)
        self.send_value()


class Roi2DRing(AbstractROI, ROI):
    _USE_BRIGHT_COLOR = True

    def __init__(self, value, parent=None):
        self._center = (0, 0)
        self._radius = value.radius
        self._width = value.width
        self._angle = value.angle
        self._angle_std = value.angle_std
        AbstractROI.__init__(self, value)
        ROI.__init__(
            self, self._center,
            (self._radius, self._radius), movable=False, parent=parent)
        self.aspectLocked = True
        self.init_roi()
        self.set_radius(self._radius)

    @property
    def value(self):
        self.parameters = self.parameters._replace(
            radius=self._radius, width=self._width, angle=self._angle,
            angle_std=self._angle_std)
        return self.parameters

    @value.setter
    def value(self, value: RoiParameters):
        if value != self.parameters:
            self.parameters = value
            self.set_radius(value.radius)
            self.set_width(value.width)
            if value.angle is not None:
                self.set_angle(value.angle)
            if value.angle_std is not None:
                self.set_angle_std(value.angle_std)

    def set_color(self, color):
        self.setPen(color)

    def set_movable(self, movable: bool):
        pass

    def set_center(self, center: tuple):
        self._center = center
        d = self._radius + self._width / 2
        pos = (center[1] - d, center[0] - d)
        self.setPos(pos)

    def set_radius(self, radius):
        self._radius = radius
        s = 2 * radius + self._width
        self.setSize((s, s))
        self.set_center(self._center)

    def set_width(self, width):
        self._width = width
        self.set_radius(self._radius)

    def set_angle(self, angle):
        self._angle = angle
        self.set_center(self._center)

    def set_angle_std(self, angle):
        self._angle_std = angle
        self.set_center(self._center)

    def paint(self, p, opt, widget):
        p.setRenderHint(QPainter.Antialiasing)
        p.setPen(self.currentPen)

        x1, y1 = 0, 0
        x2, y2 = x1 + self._width, y1 + self._width
        x3, y3 = x1 + self._width / 2, y1 + self._width / 2
        d1, d2, d3 = (2 * self._radius + self._width,
                      2 * self._radius - self._width,
                      2 * self._radius)

        # p.scale(self._radius, self._radius)
        r1 = QRectF(x1, y1, d1, d1)
        r2 = QRectF(x2, y2, d2, d2)
        r3 = QRectF(x3, y3, d3, d3)
        angle = - self._angle or 0
        angle_std = self._angle_std or 360
        a1, a2 = ((angle - angle_std / 2) * 16, angle_std * 16)
        p.drawArc(r1, a1, a2)
        p.drawArc(r2, a1, a2)
        dash_pen = QPen(self.currentPen)
        dash_pen.setStyle(Qt.DashLine)
        p.setPen(dash_pen)
        p.drawArc(r3, a1, a2)

    def getArrayRegion(self, arr, img=None, axes=(0, 1), **kwds):
        """
        Return the result of ROI.getArrayRegion()
        masked by the elliptical shape
        of the ROI. Regions outside the ellipse are set to 0.
        """
        # Note: we could use the same method as used by PolyLineROI, but this
        # implementation produces a nicer mask.
        arr = ROI.getArrayRegion(self, arr, img, axes, **kwds)
        if arr is None or arr.shape[axes[0]] == 0 or arr.shape[axes[1]] == 0:
            return arr
        w = arr.shape[axes[0]]
        h = arr.shape[axes[1]]
        ## generate an ellipsoidal mask
        mask = np.fromfunction(
            lambda x, y: (((x + 0.5) / (w / 2.) - 1) ** 2 + ((y + 0.5) / (h / 2.) - 1) ** 2) ** 0.5 < 1, (w, h))

        # reshape to match array axes
        if axes[0] > axes[1]:
            mask = mask.T
        shape = [(n if i in axes else 1) for i, n in enumerate(arr.shape)]
        mask = mask.reshape(shape)

        return arr * mask

    def shape(self):
        self.path = QPainterPath()
        self.path.addEllipse(self.boundingRect())
        return self.path


class RingParametersWidget(AbstractROI, QWidget):
    deleteClicked = pyqtSignal(object)

    @property
    def item(self):
        item = self._item()
        if not item:
            raise LookupError(f'Item is not found.')
        return item

    @property
    def control_widget(self):
        control_widget = self._control_widget()
        if not control_widget:
            raise LookupError(f'Control widget is not found.')
        return control_widget

    @property
    def value(self):
        self.parameters = self.parameters._replace(
            radius=self.radius_slider.value, width=self.width_slider.value
        )
        return self.parameters

    @value.setter
    def value(self, value):
        if value != self.parameters:
            self.radius_slider.set_value(value.radius, True)
            self.width_slider.set_value(value.width, True)

    def __init__(self,
                 model_item: QStandardItem,
                 control_widget: 'ControlWidget',
                 params: RoiParameters,
                 radius_range,
                 width_range,
                 decimals: int = 3
                 ):
        QWidget.__init__(self)
        AbstractROI.__init__(self, params)
        self._item = weakref.ref(model_item)
        self._control_widget = weakref.ref(control_widget)
        self.setup_window = None
        self.name = params.name
        self.__init_ui__(params, radius_range, width_range, decimals)
        self.init_roi()

    def set_name(self, name: str):
        super().set_name(name)
        self.name = name
        self.label.setText(name)

    def __init_ui__(self, params, radius_range, width_range, decimals):
        self.width_slider = ControlSlider('Width', width_range,
                                          params.width,
                                          self, decimals=decimals)
        self.radius_slider = ControlSlider('Radius', radius_range,
                                           params.radius,
                                           self, decimals=decimals)
        self.setup_button = RoundedPushButton(
            icon=Icon('setup_white'), radius=30,
            background_color=QColor(255, 255, 255, 100))
        self.delete_button = DeleteButton(self)
        self.setup_button.clicked.connect(self.open_setup_window)
        self.delete_button.clicked.connect(
            lambda: self.deleteClicked.emit(self.value))
        self.radius_slider.statusChanged.connect(self.on_slider_status_changed)
        self.width_slider.statusChanged.connect(self.on_slider_status_changed)

        self.radius_slider.valueChanged.connect(self.send_value)
        self.width_slider.valueChanged.connect(self.send_value)

        layout = QHBoxLayout(self)
        self.label = QLineEdit(self.name)
        self.label.textEdited.connect(self.send_name)
        self.label.setStyleSheet('QLineEdit {border: none;}')
        self.setLayout(layout)
        layout.addWidget(self.label)
        layout.addWidget(self.radius_slider)
        layout.addWidget(self.width_slider)
        layout.addWidget(self.setup_button)
        layout.addWidget(self.delete_button)

    def on_slider_status_changed(self, status):
        if status == 'show':
            self.send_active()

    def open_setup_window(self):
        RoiContextMenu(self)

    def set_active(self, *args):
        if not self._active:
            self._active = True
            self.control_widget.selectionModel().select(
                self.item.index(), QItemSelectionModel.Select)

    def set_inactive(self):
        if self._active:
            self._active = False
            self.control_widget.selectionModel().select(
                self.item.index(), QItemSelectionModel.Deselect)
            self.radius_slider.hide_and_show('hide')
            self.width_slider.hide_and_show('hide')

    def set_color(self, color):
        pass

    def set_movable(self, movable: bool):
        if not movable:
            self.radius_slider.set_fixed()
            self.width_slider.set_fixed()
        else:
            self.radius_slider.set_unfixed()
            self.width_slider.set_unfixed()

    def set_radius_bounds(self, bounds: tuple, change_value: bool = False):
        self.radius_slider.set_bounds(bounds, change_value)

    def set_width_bounds(self, bounds: tuple, change_value: bool = False):
        self.width_slider.set_bounds(bounds, change_value)


class RingSegmentParametersWidget(RingParametersWidget):
    @property
    def value(self):
        self.parameters = self.parameters._replace(
            radius=self.radius_slider.value, width=self.width_slider.value,
            angle=self.angle_slider.value, angle_std=self.angle_std_slider.value
        )
        return self.parameters

    @value.setter
    def value(self, value):
        if value != self.parameters:
            self.radius_slider.set_value(value.radius, True)
            self.width_slider.set_value(value.width, True)
            self.angle_slider.set_value(value.angle, True)
            self.angle_std_slider.set_value(value.angle_std, True)

    def __init_ui__(self, params, radius_range, width_range, decimals):
        self.width_slider = ControlSlider('Width', width_range,
                                          params.width,
                                          self, decimals=decimals)
        self.radius_slider = ControlSlider('Radius', radius_range,
                                           params.radius,
                                           self, decimals=decimals)
        self.angle_slider = ControlSlider('Angle', (0, 360),
                                          params.angle,
                                          self, decimals=decimals)
        self.angle_std_slider = ControlSlider('Angle width', (0, 360),
                                              params.angle_std,
                                              self, decimals=decimals)
        self.setup_button = RoundedPushButton(
            icon=Icon('setup_white'), radius=30,
            background_color=QColor(255, 255, 255, 100))
        self.delete_button = DeleteButton(self)
        self.setup_button.clicked.connect(self.open_setup_window)
        self.delete_button.clicked.connect(
            lambda: self.deleteClicked.emit(self.value))
        self.radius_slider.statusChanged.connect(self.on_slider_status_changed)
        self.width_slider.statusChanged.connect(self.on_slider_status_changed)
        self.angle_slider.statusChanged.connect(self.on_slider_status_changed)
        self.angle_std_slider.statusChanged.connect(self.on_slider_status_changed)

        self.radius_slider.valueChanged.connect(self.send_value)
        self.width_slider.valueChanged.connect(self.send_value)
        self.angle_slider.valueChanged.connect(self.send_value)
        self.angle_std_slider.valueChanged.connect(self.send_value)

        layout = QGridLayout(self)
        self.label = QLineEdit(self.name)
        self.label.editingFinished.connect(
            lambda: self.set_name(self.label.text())
        )
        self.label.setStyleSheet('QLineEdit {border: none;}')
        self.setLayout(layout)
        layout.addWidget(self.label, 0, 0, 2, 1)
        layout.addWidget(self.radius_slider, 0, 1)
        layout.addWidget(self.angle_slider, 0, 2)
        layout.addWidget(self.width_slider, 1, 1)
        layout.addWidget(self.angle_std_slider, 1, 2)
        layout.addWidget(self.setup_button, 0, 3, 2, 1)
        layout.addWidget(self.delete_button, 0, 4, 2, 1)

    def set_inactive(self):
        if self._active:
            self._active = False
            self.control_widget.selectionModel().select(
                self.item.index(), QItemSelectionModel.Deselect)
            self.radius_slider.hide_and_show('hide')
            self.width_slider.hide_and_show('hide')
            self.angle_slider.hide_and_show('hide')
            self.angle_std_slider.hide_and_show('hide')

    def set_movable(self, movable: bool):
        if not movable:
            self.radius_slider.set_fixed()
            self.width_slider.set_fixed()
            self.angle_slider.set_fixed()
            self.angle_std_slider.set_fixed()
        else:
            self.radius_slider.set_unfixed()
            self.width_slider.set_unfixed()
            self.angle_slider.set_unfixed()
            self.angle_std_slider.set_unfixed()


class FileWidgetRoi(AbstractROI, QObject):
    _USE_BRIGHT_COLOR = True

    @property
    def item(self):
        return self._item() if self._item else None

    def __init__(self, value: RoiParameters, item: 'PropertiesItem' = None):
        AbstractROI.__init__(self, value)
        QObject.__init__(self)
        self._item = weakref.ref(item) if item else None

    def set_item(self, item: 'PropertiesItem'):
        item.roi = self
        self._item = weakref.ref(item)

    @property
    def value(self):
        return self.parameters

    @value.setter
    def value(self, value: RoiParameters):
        self.parameters = value

    def set_color(self, color):
        if self.item:
            self.item.setData(color, Qt.ForegroundRole)

    def set_movable(self, movable: bool):
        pass

    def set_name(self, name: str):
        super().set_name(name)
        self.item.setText(name)



# -*- coding: utf-8 -*-
import logging

from .signal_connection import SignalContainer, SignalConnector, StatusChangedContainer
from .roi.roi_widgets import Roi1DAngular
from .roi.roi_containers import BasicROIContainer

from .basic_widgets import Smooth1DPlot
from ..utils import RoiParameters

logger = logging.getLogger(__name__)


class AngularProfileWidget(BasicROIContainer, Smooth1DPlot):
    def __init__(self, signal_connector: SignalConnector, parent=None):
        BasicROIContainer.__init__(self, signal_connector)
        Smooth1DPlot.__init__(self, parent)
        self._update_suggested = False
        self.current_roi_key = None
        self.bins_number = 300

    def process_signal(self, sc: SignalContainer):
        for signal in sc.segment_moved():
            if signal().key == self.current_roi_key:
                self.update_profile()
        BasicROIContainer.process_signal(self, sc)

    def _on_status_changed(self, sig: StatusChangedContainer):
        super()._on_status_changed(sig)
        selected_params = self.get_selected()
        if (
                len(selected_params) == 1 and
                self.current_roi_key != selected_params[0].key
        ):
            self._change_current_roi(selected_params[0].key)

    def _change_current_roi(self, k):
        self._remove_current_roi()
        self.current_roi_key = k
        self.roi_dict[self.current_roi_key].show()
        self.update_profile()

    def _remove_current_roi(self):
        if self.current_roi_key is not None:
            self.roi_dict[self.current_roi_key].hide()
            self.current_roi_key = None

    def _get_roi(self, params: RoiParameters):
        roi = Roi1DAngular(params)
        roi.hide()
        return roi

    def _add_item(self, roi):
        self.image_view.plot_item.addItem(roi)

    def _remove_item(self, roi):
        if roi.key == self.current_roi_key:
            self.current_roi_key = None
        self.image_view.plot_item.removeItem(roi)

    def update_profile(self):
        if any(x is None for x in
               (self.current_roi_key,
                self.image.phi,
                self.image.image,
                self.image.rr)):
            return
        roi = self.roi_dict[self.current_roi_key]
        r, w = roi.value.radius / self.image.scale, roi.value.width / self.image.scale
        r1, r2 = r - w / 2, r + w / 2
        self.x, self.y = self.image.get_angular_profile(r1, r2)
        self.plot()

    def send_value_changed(self, value: RoiParameters):
        # TODO: makes sense to switch to dependency injection scheme with one segments holder.
        sc = SignalContainer(app_node=self)
        if value.type == RoiParameters.roi_types.ring:
            value = value._replace(type=RoiParameters.roi_types.segment)
            self.roi_dict[value.key].parameters = value
            sc.type_changed(value)
        sc.segment_moved(value)
        sc.send()



# -*- coding: utf-8 -*-
from abc import abstractmethod

from PyQt5.QtCore import pyqtSignal

from .colors import COLOR_DICT
from ..signal_connection import StatusChangedContainer, SignalContainer
from ...utils import RoiParameters


class AbstractROI(object):
    value_changed = pyqtSignal(RoiParameters)
    status_changed = pyqtSignal(StatusChangedContainer)
    arbitrary_signal = pyqtSignal(SignalContainer)
    signal_by_key = pyqtSignal(tuple)

    _USE_BRIGHT_COLOR = False

    @property
    def color(self):
        # TODO: set colors independently for better customization?
        value = self.parameters
        if value.movable and self._active:
            return COLOR_DICT['active_bright'] if self._USE_BRIGHT_COLOR else COLOR_DICT['active']
        if value.movable and not self._active:
            return COLOR_DICT['inactive_bright'] if self._USE_BRIGHT_COLOR else COLOR_DICT['inactive']
        if not value.movable and self._active:
            return COLOR_DICT['fixed_active_bright'] if self._USE_BRIGHT_COLOR else COLOR_DICT['fixed_active']
        if not value.movable and not self._active:
            return COLOR_DICT['fixed_bright'] if self._USE_BRIGHT_COLOR else COLOR_DICT['fixed']

    def __init__(self, value: RoiParameters):
        self.parameters = value
        self.key = value.key
        self._active = False

    def init_roi(self):
        self.set_inactive()
        self.set_value(self.parameters)
        if not self.parameters.movable:
            self.set_fixed()

    @property
    def active(self):
        return self._active

    @property
    @abstractmethod
    def value(self):
        """
        Value is an instance of RoiParameters class.
        :return:
        """
        pass

    @value.setter
    @abstractmethod
    def value(self, value: RoiParameters):
        """
        Value is an instance of RoiParameters class.
        :param value: RoiParameters
        :return:
        """
        pass

    def set_value(self, value):
        if self.value != value:
            self.value = value

    def set_active(self):
        self._active = True
        self.set_color(self.color)

    def set_inactive(self):
        self._active = False
        self.set_color(self.color)

    def change_active(self, change_others: bool = True):
        if self._active:
            self.send_inactive()
        else:
            self.send_active(change_others)

    def set_fixed(self):
        self.parameters = self.parameters._replace(movable=False)
        self.set_movable(False)
        self.set_color(self.color)

    def set_unfixed(self):
        self.parameters = self.parameters._replace(movable=True)
        self.set_movable(True)
        self.set_color(self.color)

    @abstractmethod
    def set_color(self, color):
        pass

    @abstractmethod
    def set_movable(self, movable: bool):
        pass

    def send_value(self):
        self.value_changed.emit(self.value)

    def send_active(self, change_others: bool = True):
        self.status_changed.emit(StatusChangedContainer(
            [self.key], True, change_others))

    def send_name(self, name: str, **kwargs):
        self.set_name(name)
        self.arbitrary_signal.emit(SignalContainer().name_changed(self.value, **kwargs))

    def set_name(self, name: str):
        self.parameters = self.parameters._replace(name=name)
        # logger.debug(self.value)

    def send_inactive(self):
        self.status_changed.emit(StatusChangedContainer(
            [self.key], False))

    def send_signal(self, signal_container: SignalContainer):
        self.arbitrary_signal.emit(signal_container)

    def __repr__(self):
        return str(self.parameters)



import logging

import numpy as np

from PyQt5.QtWidgets import (QWidget, QHBoxLayout,
                             QTreeView, QLabel)
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QStandardItem, QStandardItemModel

from .basic_widgets import (RoundedPushButton,
                            DeleteButton)
from .roi.roi_widgets import RingParametersWidget, RingSegmentParametersWidget
from .roi.roi_containers import BasicROIContainer
from .signal_connection import (SignalConnector,
                                SignalContainer)
from ..utils import Icon, RoiParameters

logger = logging.getLogger(__name__)


class ControlWidget(BasicROIContainer, QTreeView):

    def __init__(self, signal_connector: SignalConnector, parent=None):
        BasicROIContainer.__init__(self, signal_connector)
        QTreeView.__init__(self, parent)
        self._model = QStandardItemModel()
        self._model.setHorizontalHeaderLabels([''])
        self._model.setRowCount(0)
        self.setModel(self._model)
        self.selectionModel().currentChanged.connect(self.on_clicked)

        # the following variables will be wisely redefined when geometry is changed
        self._radius_bounds = np.array([0, 2000])
        self._width_bounds = np.array([0, 200])

        # TODO: common place to store init parameters for new rois ?
        self._default_ring_parameters = dict(
            radius=10, width=1, angle=180, angle_std=360,
            type=RoiParameters.roi_types.ring)
        self._default_segment_parameters = dict(
            radius=10, width=1, angle=180, angle_std=180,
            type=RoiParameters.roi_types.segment)

        self.__init_ui__()
        self.show()

    def __init_ui__(self):
        self.add_ring_button = RoundedPushButton(
            text='Add', icon=Icon('add'), radius=30)
        self.add_ring_button.clicked.connect(self.emit_create_ring)
        self.delete_all_rings_button = DeleteButton(text='Delete all rings?')
        self.delete_all_rings_button.clicked.connect(self.emit_delete_all_rings)

        self.add_arc_button = RoundedPushButton(
            text='Add', icon=Icon('add'), radius=30)
        self.add_arc_button.clicked.connect(self.emit_create_arc)
        self.delete_all_arcs_button = DeleteButton(text='Delete all segments?')
        self.delete_all_arcs_button.clicked.connect(self.emit_delete_all_arcs)

        self.ring_item = QStandardItem()
        ring_layout = self._get_header_layout(self.ring_item, 'Rings')
        ring_layout.addWidget(self.add_ring_button, alignment=Qt.AlignRight)
        ring_layout.addWidget(self.delete_all_rings_button, alignment=Qt.AlignRight)

        self.arc_item = QStandardItem()
        arc_layout = self._get_header_layout(self.arc_item, 'Ring Segments')
        arc_layout.addWidget(self.add_arc_button, alignment=Qt.AlignRight)
        arc_layout.addWidget(self.delete_all_arcs_button, alignment=Qt.AlignRight)

        self.setGeometry(500, 500, 500, 500)
        self.setColumnWidth(0, 250)
        self.header().resizeSection(0, 100)

    def process_signal(self, s: SignalContainer):
        for _ in s.geometry_changed():
            self._on_geometry_changed()
        super().process_signal(s)

    def _on_geometry_changed(self):
        if self.image.rr is not None:
            self._radius_bounds = np.array([self.image.rr.min() * self.image.scale,
                                            self.image.rr.max() * self.image.scale])
            self._width_bounds = np.array([0, self.image.rr.max() * self.image.scale / 10])
            self._default_ring_parameters['radius'] = self._radius_bounds.sum() / 2
            self._default_ring_parameters['width'] = self._radius_bounds.sum() / 10

    def _get_header_layout(self, item: QStandardItem, label: str):
        header_widget = QWidget(self)
        layout = QHBoxLayout()
        header_widget.setLayout(layout)
        label_widget = QLabel(label)
        layout.addWidget(label_widget, alignment=Qt.AlignLeft)
        layout.addStretch(1)
        self._model.appendRow(item)
        self.setIndexWidget(item.index(), header_widget)
        return layout

    def on_clicked(self, index):
        item = self._model.itemFromIndex(index)
        key = item.data()
        if key is not None:
            selected_roi = self.roi_dict[key]
            selected_roi.send_active()

    def emit_create_ring(self, *args):
        params = RoiParameters(**self._default_ring_parameters,
                               name=f'ring {self.ring_item.rowCount()}')
        self.__class__.emit_create_segment(self, params)

    def emit_create_arc(self, *args):
        params = RoiParameters(**self._default_segment_parameters,
                               name=f'ring segment {self.arc_item.rowCount()}')
        self.__class__.emit_create_segment(self, params)

    def emit_delete_all_of_type(self, roi_type: int):
        sc = SignalContainer()
        for roi in self.roi_dict.values():
            if roi.parameters.type == roi_type:
                sc.segment_deleted(roi)
        self.signal_connector.emit_upward(sc)

    def emit_delete_all_rings(self):
        return self.emit_delete_all_of_type(RoiParameters.roi_types.ring)

    def emit_delete_all_arcs(self):
        return self.emit_delete_all_of_type(RoiParameters.roi_types.segment)

    def _get_roi(self, params: RoiParameters):
        new_ring_item = QStandardItem()
        new_ring_item.setData(params.key)

        if params.type == RoiParameters.roi_types.ring:
            new_roi = RingParametersWidget(
                new_ring_item, self, params,
                tuple(self._radius_bounds),  tuple(self._width_bounds))
            self.ring_item.appendRow(new_ring_item)
        elif params.type == RoiParameters.roi_types.segment:
            new_roi = RingSegmentParametersWidget(
                new_ring_item, self, params,
                tuple(self._radius_bounds), tuple(self._width_bounds))
            self.arc_item.appendRow(new_ring_item)
        self._model.layoutChanged.emit()
        return new_roi

    def _add_item(self, roi: RingParametersWidget):
        self.setIndexWidget(roi.item.index(), roi)
        self.setExpanded(roi.item.parent().index(), True)
        roi.deleteClicked.connect(self.emit_delete_segment)

    def _remove_item(self, roi: RingParametersWidget):
        ring_item = roi.item
        roi.item.parent().removeRow(ring_item.row())
        self._model.layoutChanged.emit()

    def on_type_changed(self, value: RoiParameters):
        self.delete_roi(self.roi_dict[value.key])
        self.add_roi(value)

    def _on_scale_changed(self):
        self._radius_bounds = self._radius_bounds * self.image.scale_change
        self._width_bounds = self._width_bounds * self.image.scale_change
        for v in self.roi_dict.values():
            v.set_radius_bounds(self._radius_bounds)
            v.set_width_bounds(self._width_bounds)



# -*- coding: utf-8 -*-
from abc import abstractmethod

from ..exceptions import KeySignalNameError
from ..signal_connection import (AppNode, SignalConnector, SignalContainer,
                                 StatusChangedContainer)
from ...utils import RoiParameters


class AbstractROIContainer(AppNode):
    def __init__(self, signal_connector: SignalConnector):
        AppNode.__init__(self, signal_connector)
        self.signal_connector.downwardSignal.connect(self.process_signal)
        self.roi_dict = dict()

    def process_signal(self, s: SignalContainer):
        for _ in s.scale_changed():
            self._on_scale_changed()
        for signal in s.segment_created():
            self.add_roi(signal())
        for signal in s.segment_moved():
            self.roi_dict[signal().key].value = signal()
        for signal in s.segment_deleted():
            self.delete_roi(signal())
        for signal in s.segment_fixed():
            self.roi_dict[signal().key].set_fixed()
        for signal in s.segment_unfixed():
            self.roi_dict[signal().key].set_unfixed()
        for signal in s.status_changed():
            self._on_status_changed(signal())
        for signal in s.name_changed():
            self.roi_dict[signal().key].set_name(signal().name)
        for signal in s.type_changed():
            self.on_type_changed(signal())

    def _on_scale_changed(self):
        pass

    @abstractmethod
    def _get_roi(self, params: RoiParameters) -> 'AbstractROI':
        pass

    @abstractmethod
    def _add_item(self, roi: 'AbstractROI'):
        pass

    @abstractmethod
    def _remove_item(self, roi: 'AbstractROI'):
        pass

    def _on_status_changed(self, sig: StatusChangedContainer):
        if not sig.status:
            for k in sig.keys:
                self.roi_dict[k].set_inactive()
        else:
            for k in sig.keys:
                self.roi_dict[k].set_active()

    def emit_create_segment(self, params: RoiParameters):
        self.signal_connector.emit_upward(
            SignalContainer().segment_created(params))

    def emit_delete_segment(self, params: RoiParameters):
        self.signal_connector.emit_upward(
            SignalContainer().segment_deleted(params))

    def emit_status_changed(self, sig: StatusChangedContainer):
        if (not sig.status and
                len(sig.keys) == 1 and
                set([roi.key for roi in self.get_selected()]).difference(sig.keys)):
            sig = sig._replace(status=True, change_others=True)
        SignalContainer(app_node=self).status_changed(sig).send()

    def add_roi(self, params: RoiParameters):
        roi = self._get_roi(params)
        if roi:
            roi.value_changed.connect(self.send_value_changed)
            roi.status_changed.connect(self.emit_status_changed)
            roi.arbitrary_signal.connect(self.signal_connector.emit_upward)
            self.roi_dict[params.key] = roi
            self._add_item(roi)
            return roi

    def delete_roi(self, params: RoiParameters):
        roi = self.roi_dict.pop(params.key)
        self._remove_item(roi)
        roi.deleteLater()

    def get_selected(self):
        return [roi.parameters for roi in self.roi_dict.values() if roi.active]

    def send_value_changed(self, value: RoiParameters):
        SignalContainer(app_node=self).segment_moved(value).send()

    def on_type_changed(self, value: RoiParameters):
        self.roi_dict[value.key].value = value


class BasicROIContainer(AbstractROIContainer):
    def add_roi(self, params: RoiParameters):
        roi = AbstractROIContainer.add_roi(self, params)
        if roi:
            roi.signal_by_key.connect(self.resend_key_signal)
        return roi

    def resend_key_signal(self, key: tuple):
        key, args = key[0], key[1:]
        method = getattr(self, key, None)
        if not method:
            raise KeySignalNameError()
        method(*args)

    def fix_selected(self):
        sc = SignalContainer(app_node=self)
        for selected in self.get_selected():
            sc.segment_fixed(selected)
        sc.send()

    def unfix_selected(self):
        sc = SignalContainer(app_node=self)
        for selected in self.get_selected():
            sc.segment_unfixed(selected)
        sc.send()

    def fix_all(self):
        sc = SignalContainer(app_node=self)
        for roi in self.roi_dict.values():
            sc.segment_fixed(roi.parameters)
        sc.send()

    def unfix_all(self):
        sc = SignalContainer(app_node=self)
        for roi in self.roi_dict.values():
            sc.segment_unfixed(roi.parameters)
        sc.send()

    def delete_selected(self):
        sc = SignalContainer(app_node=self)
        for selected in self.get_selected():
            sc.segment_deleted(selected)
        sc.send()

    def select_all(self):
        sc = SignalContainer(app_node=self)
        keys = list(self.roi_dict.keys())
        sc.status_changed(StatusChangedContainer(keys, True, False))
        sc.send()

    def unselect_all(self):
        sc = SignalContainer(app_node=self)
        keys = list(self.roi_dict.keys())
        sc.status_changed(StatusChangedContainer(keys, False, False))
        sc.send()

    def change_roi_type(self, value: RoiParameters):
        self.on_type_changed(value)
        SignalContainer(app_node=self).type_changed(value).send()


import logging
from typing import NamedTuple

import numpy as np
import cv2

from .modes import get_mode
from .parameters_widget import get_interpolation_parameters

logger = logging.getLogger(__name__)


class Interpolation(object):
    """
    This class is a singleton and it contains main functionality needed for 2d polar interpolation.
    Its instance is held by ..global_context.Image class so that other widgets could
    get access to it (for instance, for fast angular profile calculation).
    """

    def __init__(self):
        self._interpolation_geometry = None
        self._image = None
        params = get_interpolation_parameters()
        self._r_size = params.get('r_size', None)
        self._phi_size = params.get('phi_size', None)
        mode_name = params.get('mode', '')
        self._algorithm = get_mode(mode_name)

    @property
    def interpolation_geometry(self) -> 'InterpolationGeometry' or None:
        return self._interpolation_geometry

    @property
    def image(self) -> np.ndarray or None:
        return self._image

    @property
    def r_size(self):
        return self._r_size

    @property
    def phi_size(self):
        return self._phi_size

    @property
    def algorithm_flag(self) -> int:
        return self._algorithm.flag

    @property
    def algorithm_name(self) -> str:
        return self._algorithm.name

    @property
    def r_axis(self) -> np.ndarray or None:
        try:
            return self._interpolation_geometry.r
        except AttributeError:  # should be faster than checking if not None
            return

    @property
    def phi_axis(self) -> np.ndarray or None:
        try:
            return self._interpolation_geometry.p
        except AttributeError:
            return

    def set_algorithm(self, mode: 'Mode' or str):
        if isinstance(mode, str):
            mode = get_mode(mode)
        if not mode:
            return
        self._algorithm = mode

    def set_geometry(self, geometry: 'Geometry', r_size: int = None, phi_size: int = None):
        self.set_shape(r_size, phi_size)
        self._interpolation_geometry = InterpolationGeometry.get(
            geometry, self._r_size, self._phi_size)

    def set_shape(self, r_size: int = None, phi_size: int = None):
        if r_size:  # anyway should be nonzero
            self._r_size = r_size
        if phi_size:
            self._phi_size = phi_size

    def set_parameters(self, parameters: dict):
        self.set_shape(parameters.get('r_size', None), parameters.get('phi_size', None))
        self.set_algorithm(parameters.get('mode', None))

    def interpolate(self, image: np.ndarray) -> np.ndarray or None:
        if any(x is None for x in (self.interpolation_geometry, image)):
            return
        xx = self.interpolation_geometry.xx.astype(np.float32)
        yy = self.interpolation_geometry.yy.astype(np.float32)
        try:
            logger.info(f'Calculating interpolation.')
            self._image = cv2.remap(image.astype(np.float32), xx, yy,
                                    interpolation=self.algorithm_flag)
            logger.info(f'Interpolation is calculated.')
            return self._image
        except cv2.error as err:
            logger.exception(err)
            return

    def get_angular_profile(self, r1: int, r2: int):
        if self.image is None:
            return
        max_ind = self.r_size - 1
        r1, r2 = max(min((r1, r2)), 0), min(max((r1, r2)), max_ind)
        if r1 > max_ind or r2 < 0:
            return
        try:
            return self.image[:, r1:r2].sum(axis=1)
        except Exception as err:
            logger.exception(err)
            return


class InterpolationGeometry(NamedTuple):
    """
    Interpolation geometry container used for polar interpolation.
    Should only be initialized by the class method '.get()'.
    Fields:
        r: np.ndarray - 1d radius axis for interpolated image
        p: np.ndarray - 1d angular axis [rad] for interpolated image
        xx: np.ndarray - 2d map1 for cv2 remap function
        yy: np.ndarray - 2d map2 for cv2 remap function
    """
    r: np.ndarray
    p: np.ndarray
    xx: np.ndarray
    yy: np.ndarray

    @classmethod
    def get(cls, geometry: 'Geometry', r_size: int, phi_size: int):
        rr = geometry.rr
        phi = geometry.phi
        center = geometry.beam_center
        if any(x is None for x in (rr, phi, center, r_size, phi_size)):
            return
        r = np.linspace(rr.min(), rr.max(), r_size)
        r_matrix = r[np.newaxis, :].repeat(phi_size, axis=0)
        p = np.linspace(phi.min(), phi.max(), phi_size)
        p_matrix = p[:, np.newaxis].repeat(r_size, axis=1)
        xx = r_matrix * np.cos(p_matrix) + center[1]
        yy = r_matrix * np.sin(p_matrix) + center[0]
        return cls(r=r, p=p, xx=xx, yy=yy)



# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Iterable, List

import numpy as np
import cv2

from .read_edf import read_edf_from_file

_PROJECT_DIR = Path(__file__).parents[1]

_TIFF_PATHS = list(_PROJECT_DIR.rglob('*.tiff'))
_EDF_PATHS = list(_PROJECT_DIR.rglob('*.edf'))


def get_image(num: int = 0, *, random: bool = False,
              normalize: bool = True) -> np.array:
    if random:
        filepath = np.random.choice(_TIFF_PATHS, 1)[0]
    else:
        try:
            filepath = _TIFF_PATHS[num]
        except IndexError:
            raise FileNotFoundError()
    return get_image_from_path(filepath, normalize=normalize)


def get_images(nums: Iterable or int, normalize: bool = True) -> List[np.array]:
    if isinstance(nums, int):
        image_paths = np.random.choice(_TIFF_PATHS, nums, replace=False)
    else:
        image_paths = [_TIFF_PATHS[n] for n in nums]
    return [get_image_from_path(path, normalize) for path in image_paths]


def get_image_from_path(filepath, normalize: bool = True) -> np.array:
    filepath = str(filepath)
    if filepath.endswith('.edf'):
        image = read_edf_from_file(filepath)[0]
    else:
        image = np.flip(cv2.imread(filepath, cv2.IMREAD_GRAYSCALE), 0)
    if normalize:
        image = normalize_image(image)
    return image


def get_raw_image(num: int = 0, *, random: bool = False):
    if random:
        num = np.random.choice(list(range(len(_EDF_PATHS))), 1)[0]
    try:
        return read_edf_from_file(str(_EDF_PATHS[num]))[0]
    except IndexError:
        raise FileNotFoundError()


def normalize_image(image: np.ndarray):
    # TODO: move to a standalone optional functionality with parameter sigma_factor
    # ( hidden vertical slider on the widget on the right ? )
    m, s = image.mean(), image.std() * 2
    image = np.clip(image, m - s, m + s)
    image = (image - image.min()) / image.max()
    return image




# -*- coding: utf-8 -*-
import logging
from abc import abstractmethod
from typing import NamedTuple

import numpy as np
from scipy.ndimage import gaussian_filter1d

from PyQt5.QtWidgets import (QWidget, QMainWindow, QFrame,
                             QSlider, QLineEdit, QHBoxLayout,
                             QLabel, QAbstractButton,
                             QGraphicsDropShadowEffect,
                             QWidgetAction, QMenu, QFormLayout,
                             QPushButton, QRadioButton)
from PyQt5.QtGui import (QIntValidator, QPainter, QBrush, QColor)
from PyQt5.QtCore import Qt, pyqtSignal, QRect

from pyqtgraph import (GraphicsLayoutWidget, setConfigOptions,
                       ImageItem, HistogramLUTItem)

from ..config import read_config, save_config
from ..utils import (color_animation, QIcon, Icon,
                     center_widget, validate_scientific_value)

logger = logging.getLogger(__name__)


class Custom1DPlot(GraphicsLayoutWidget):
    def __init__(self, *args, parent=None):
        super(Custom1DPlot, self).__init__(parent)
        self.plot_item = self.addPlot()
        self.plot_item.setMenuEnabled(False)
        self.plot = self.plot_item.plot(*args)

    def set_data(self, *args):
        self.plot.setData(*args)

    def clear_data(self):
        self.plot_item.clearPlots()

    def set_x(self, x):
        x = np.array(x)
        y = self.plot.yData
        if y.shape != x.shape:
            return
        self.plot.setData(x, y)


class Smooth1DPlot(QMainWindow):
    _MaximumSliderWidth = 200
    _MaximumSliderHeight = 30

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        self._y = value
        self.update_smoothed_y()

    @property
    def smoothed_y(self):
        return self._smoothed_y

    @property
    def x(self):
        if self._x is None and self._y is not None:
            return np.arange(self._y.size)
        else:
            return self._x

    @x.setter
    def x(self, value):
        self._x = value

    def update_smoothed_y(self):
        y = self.y
        if isinstance(y, np.ndarray):
            if self.sigma:
                self._smoothed_y = gaussian_filter1d(y, self.sigma)
            else:
                self._smoothed_y = y

    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent=parent)
        self.image_view = Custom1DPlot()
        self.setCentralWidget(self.image_view)
        self.sigma = 0
        self._y = None
        self._smoothed_y = None
        self._x = None
        self.__init_toolbars__()

    def __init_toolbars__(self):
        param_toolbar = self.addToolBar('Parameters')
        param_toolbar.setStyleSheet('background-color: black;')

        self.__init_sigma_slider__(param_toolbar)

    def __init_sigma_slider__(self, toolbar):
        sigma_slider = AnimatedSlider('𝞼', (0, 10), self.sigma,
                                      decimals=2)
        sigma_slider.setMaximumWidth(self._MaximumSliderWidth)
        sigma_slider.setMaximumHeight(self._MaximumSliderHeight)
        sigma_slider.valueChanged.connect(self.update_sigma)
        sigma_slider.setStyleSheet('background-color: white;')
        sigma_slider.shadow.setColor(QColor('blue'))
        self.sigma_slider = sigma_slider
        frame = QFrame()
        layout = QHBoxLayout()
        frame.setLayout(layout)
        frame.setGeometry(0, 0, self._MaximumSliderWidth, toolbar.height() * 0.9)
        layout.addWidget(sigma_slider, alignment=Qt.AlignHCenter)
        toolbar.addWidget(frame)

    def set_sigma(self, value: float):
        self.update_sigma(value)
        self.sigma_slider.set_value(value, change_bounds=True)

    def update_sigma(self, value: float):
        self.sigma = value
        self.update_smoothed_y()
        self.plot()

    def plot(self):
        self.image_view.set_data(self.x, self.smoothed_y)


class CustomImageViewer(GraphicsLayoutWidget):

    def __init__(self, parent=None, **kwargs):
        setConfigOptions(imageAxisOrder='row-major')
        super(CustomImageViewer, self).__init__(parent)
        self._scale = (1., 1.)
        self._center = (0, 0)

        self.__init_ui__()

    def __init_ui__(self):
        self.setWindowTitle('Image Viewer')
        self.image_plot = self.addPlot()
        self.image_plot.vb.setAspectLocked()
        self.image_plot.vb.invertY()
        self.image_item = ImageItem()
        self.image_plot.addItem(self.image_item)
        self.hist = HistogramLUTItem()
        self.hist.setImageItem(self.image_item)
        self.addItem(self.hist)

    def set_data(self, data, change_limits: bool = True, reset_axes: bool = False):
        if data is None:
            return
        self.image_item.setImage(data, change_limits)
        if change_limits:
            self.hist.setLevels(data.min(), data.max())
        if reset_axes:
            self.image_item.resetTransform()
        self.set_default_range()

    def set_default_range(self):
        axes = self.get_axes()
        self.image_plot.setRange(xRange=axes[0], yRange=axes[1])

    def set_auto_range(self):
        self.image_plot.autoRange()

    def set_levels(self, levels=None):
        if levels:
            self.hist.setLevels(levels[0], levels[1])
        else:
            self.hist.setLevels(self.image_item.image.min(),
                                self.image_item.image.max())

    def get_levels(self):
        return self.hist.getLevels()

    def set_center(self, center: tuple, pixel_units: bool = True):
        if not pixel_units:
            scale = self.get_scale()
            center = (center[0] / scale[0], center[1] / scale[1])
        if self._center != (0, 0) or self._scale != (1., 1.):
            self.image_item.resetTransform()
            self.image_item.scale(*self._scale)
        self.image_item.translate(- center[0], - center[1])
        self._center = center
        self.set_default_range()

    def set_scale(self, scale: float or tuple):
        if isinstance(scale, float) or isinstance(scale, int):
            scale = (scale, scale)
        if self._center != (0, 0) or self._scale != (1., 1.):
            self.image_item.resetTransform()
        self.image_item.scale(*scale)
        if self._center != (0, 0):
            self.image_item.translate(- self._center[0], - self._center[1])
        self._scale = scale
        self.set_default_range()

    def get_scale(self):
        # scale property is occupied by Qt superclass.
        return self._scale

    def get_center(self):
        return self._center

    def set_x_axis(self, x_min, x_max):
        self._set_axis(x_min, x_max, 0)
        self.set_default_range()

    def set_y_axis(self, y_min, y_max):
        self._set_axis(y_min, y_max, 1)
        self.set_default_range()

    def _set_axis(self, min_: float, max_: float, axis_ind: int):
        shape = self.image_item.image.shape
        scale = np.array(self._scale)
        scale[axis_ind] = (max_ - min_) / shape[axis_ind]
        center = np.array(self._center)
        center[axis_ind] = - min_ / scale[axis_ind]
        if self._center != (0, 0) or self._scale != (1., 1.):
            self.image_item.resetTransform()
        self.image_item.scale(scale[0], scale[1])
        self.image_item.translate(- center[0], - center[1])
        self._scale = tuple(scale)
        self._center = tuple(center)

    def get_axes(self):
        shape = np.array(self.image_item.image.shape)
        scale = np.array(self._scale)
        min_ = - np.array(self._center) * scale
        max_ = min_ + shape * scale
        return (min_[0], max_[0]), (min_[1], max_[1])


class DoubleSlider(QSlider):
    valueChangedByHand = pyqtSignal(float)

    def __init__(self, *args, decimals: int = 5):
        super().__init__(*args)
        self.decimals = decimals
        self._max_int = 10 ** self.decimals
        self._pressed = False
        self.sliderPressed.connect(self._set_pressed)
        self.sliderReleased.connect(self._set_released)
        self.sliderReleased.connect(self.emit_value)
        self.valueChanged.connect(self._check_and_emit)

        super().setMinimum(0)
        super().setMaximum(self._max_int)

        self._min_value = 0.0
        self._max_value = 1.0

    def emit_value(self, *args):
        self.valueChangedByHand.emit(self.value())

    def _set_pressed(self, *args):
        self._pressed = True

    def _set_released(self, *args):
        self._pressed = False

    def _check_and_emit(self, *args):
        if self._pressed:
            self.emit_value()

    def set_decimals(self, decimals):
        self.decimals = decimals
        value = self.value()
        self._max_int = 10 ** self.decimals
        super().setMinimum(0)
        super().setMaximum(self._max_int)
        self.setRange(self._min_value, self._max_value)
        self.setValue(value)

    def _update_max_int(self):
        self._max_int = self._value_range * (10 ** self.decimals)
        if self._max_int <= 0:
            self._max_int = 1
        super().setMaximum(self._max_int)

    @property
    def _value_range(self):
        return self._max_value - self._min_value

    def _real_to_view(self, value):
        try:
            return int((value - self._min_value) / self._value_range * self._max_int)
        except ZeroDivisionError:
            return 0

    def _view_to_real(self, value):
        return value / self._max_int * self._value_range + self._min_value

    def value(self):
        return self._view_to_real(super().value())

    def setValue(self, value):
        super().setValue(self._real_to_view(value))

    def setMinimum(self, value):
        if value > self._max_value:
            raise ValueError("Minimum limit cannot be higher than maximum")
        real_value = self.value()
        self._min_value = value
        self._update_max_int()
        self.setValue(real_value)
        if real_value < value:
            self.emit_value()

    def setMaximum(self, value):
        if value < self._min_value:
            raise ValueError("Minimum limit cannot be higher than maximum")

        real_value = self.value()
        self._max_value = value
        self._update_max_int()
        self.setValue(real_value)
        if real_value > value:
            self.emit_value()

    def setRange(self, p_int, p_int_1):
        real_value = self.value()
        self._min_value = p_int
        self._max_value = p_int_1
        self._update_max_int()
        self.setValue(real_value)

    def minimum(self):
        return self._min_value

    def maximum(self):
        return self._max_value


class RoundedPushButton(QAbstractButton):
    def __init__(self,
                 parent=None, *,
                 text: str = '',
                 icon: QIcon = None,
                 radius: float = 20,
                 background_color: QColor or str = 'white'):
        self._radius = radius
        self._text = text
        self._icon = icon
        if isinstance(background_color, str):
            if background_color == 'transparent':
                background_color = QColor(0, 0, 0, 0)
            else:
                background_color = QColor(background_color)
        self._background_color = background_color
        super().__init__(parent)
        self.__init_shadow__()
        self._pressed = False
        if self._icon:
            self.setFixedSize(self._radius, self._radius)

    def __init_shadow__(self):
        self.shadow = QGraphicsDropShadowEffect()
        self.shadow.setBlurRadius(self._radius)
        self.shadow.setXOffset(3)
        self.shadow.setYOffset(3)
        self.setGraphicsEffect(self.shadow)

    def _get_painter(self):
        painter = QPainter(self)
        brush = QBrush()
        brush.setColor(self._background_color)
        brush.setStyle(Qt.SolidPattern)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)
        painter.setRenderHint(QPainter.Antialiasing)
        return painter

    def paintEvent(self, QPaintEvent):
        painter = self._get_painter()
        if self._pressed:
            rect = QRect(0, 3, painter.device().width(),
                         painter.device().height() * 0.9)
        else:
            rect = QRect(0, 0, painter.device().width(),
                         painter.device().height())
        painter.drawRoundedRect(rect, self._radius, self._radius)
        if self._icon:
            pixmap = self._icon.pixmap(self.size())
            painter.drawPixmap(0, 0, pixmap)

    def mousePressEvent(self, QMouseEvent):
        self._pressed = True
        super().mousePressEvent(QMouseEvent)
        self.shadow.setEnabled(False)

    def mouseReleaseEvent(self, QMouseEvent):
        self._pressed = False
        super().mouseReleaseEvent(QMouseEvent)
        self.shadow.setEnabled(True)
        self.shadow.setBlurRadius(self._radius)


class ConfirmButton(QWidget):
    clicked = pyqtSignal()

    def __init__(self, icon: QIcon, parent=None, radius: float = 30,
                 smaller_radius: float = 25,
                 text: str = ''):
        QWidget.__init__(self, parent)
        self._icon = icon
        self._text = text
        self._status = None
        self.__init_ui__(radius, smaller_radius)
        self.hide_and_show('icon')

    def __init_ui__(self, radius, smaller_radius):
        layout = QHBoxLayout(self)
        self.setLayout(layout)

        self.icon_widget = RoundedPushButton(self, icon=self._icon,
                                             radius=radius)
        self.icon_widget.clicked.connect(self.on_clicked)
        layout.addWidget(self.icon_widget)

        self.label_widget = QLabel(self._text)
        self.confirm = RoundedPushButton(
            icon=Icon('confirm'), radius=smaller_radius)
        self.confirm.clicked.connect(self.on_clicked)
        self.confirm.clicked.connect(self.clicked.emit)
        self.decline = RoundedPushButton(
            icon=Icon('delete'), radius=smaller_radius)
        self.decline.clicked.connect(self.on_clicked)
        layout.addWidget(self.label_widget)
        layout.addWidget(self.confirm)
        layout.addWidget(self.decline)

    def on_clicked(self):
        self.hide_and_show()

    def hide_and_show(self, new_status: str = None):
        if new_status in 'icon question'.split():
            self._status = new_status
        else:
            self._status = 'icon' if self._status == 'question' else 'question'

        if self._status == 'icon':
            self.label_widget.hide()
            self.confirm.hide()
            self.decline.hide()
            self.icon_widget.show()
        else:
            self.label_widget.show()
            self.confirm.show()
            self.decline.show()
            self.icon_widget.hide()


class DeleteButton(ConfirmButton):
    def __init__(self, parent=None, radius: float = 30,
                 smaller_radius: float = 25, text: str = 'Delete?'):
        ConfirmButton.__init__(self, Icon('delete'), parent,
                               radius, smaller_radius, text=text)


class AnimatedSlider(RoundedPushButton):
    valueChanged = pyqtSignal(float)
    statusChanged = pyqtSignal(str)
    _EditMaximumWidth = 80
    _Height = 40
    _padding = 50

    @property
    def value(self):
        return self.slider.value()

    def __init__(self,
                 name: str,
                 bounds: tuple = (0, 1),
                 value: float = 0,
                 parent=None,
                 orientation=Qt.Horizontal,
                 decimals: int = 0,
                 hide: bool = True,
                 context_menu_enabled: bool = True,
                 disable_changing_status: bool = False,
                 min_max_bounds: tuple = (-10000, 10000)):

        super(AnimatedSlider, self).__init__(parent)
        self.context_menu_enabled = context_menu_enabled
        self.name = name
        self.bounds = bounds
        self._decimals = decimals
        self.init_value = value
        self.__init_ui__(orientation)
        self.clicked.connect(self.on_clicked)
        self._status = 'show'
        self._min_max_bounds = min_max_bounds
        self._disable_changing_status = disable_changing_status
        if hide:
            self.hide_and_show('hide')

    def set_min_max_bounds(self, bounds: tuple):
        if bounds[0] < bounds[1]:
            self._min_max_bounds = bounds

    def on_clicked(self):
        self.hide_and_show()
        self.statusChanged.emit(self._status)

    def contextMenuEvent(self, *args, **kwargs):
        if not self.context_menu_enabled:
            return
        event = args[0]
        menu = QMenu()
        menu.setWindowFlags(menu.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
        menu.setAttribute(Qt.WA_TranslucentBackground)
        menu.setStyleSheet("QMenu{background:rgba(255, 255, 255, 0%);}")

        decimals_slider = AnimatedSlider(
            'Decimals', (0, 6), self._decimals, hide=False,
            context_menu_enabled=False,
            disable_changing_status=True
        )
        min_slider = AnimatedSlider(
            'Min value', (self._min_max_bounds[0], self.slider.maximum()),
            self.slider.minimum(), hide=False,
            context_menu_enabled=False, disable_changing_status=True)
        max_slider = AnimatedSlider(
            'Max value', (self.slider.minimum(), self._min_max_bounds[1]),
            self.slider.maximum(), hide=False,
            context_menu_enabled=False, disable_changing_status=True)
        decimals_slider.valueChanged.connect(self.set_decimals)
        min_slider.valueChanged.connect(self.slider.setMinimum)
        min_slider.valueChanged.connect(max_slider.slider.setMinimum)
        max_slider.valueChanged.connect(self.slider.setMaximum)

        # TODO: _x - 1e-45 looks like the worst decision ever and should be well tested or better changed.
        max_slider.valueChanged.connect(
            lambda x: min_slider.slider.setMaximum(x - 1e-45))

        for w in [decimals_slider, min_slider, max_slider]:
            w_action = QWidgetAction(self)
            w_action.setDefaultWidget(w)
            menu.addAction(w_action)
        menu.exec_(self.mapToGlobal(event.pos()))

    def set_decimals(self, value):
        try:
            value = int(value)
        except ValueError as err:
            logger.exception(err)
            return
        self.slider.set_decimals(value)
        self._decimals = value
        self.editValue.setText(self._get_str_value())
        self._update_label_text()

    def set_bounds(self, bounds: tuple, change_value: bool = False):
        try:
            self.slider.setRange(*bounds)
        except ValueError as err:
            if not change_value:
                logger.exception(err)
            else:
                self.slider.setValue(bounds[0])
                self.slider.setRange(*bounds)

    def hide_and_show(self, new_status: str = None):
        if self._disable_changing_status:
            return
        if new_status in 'show hide'.split():
            self._status = new_status
        else:
            self._status = 'hide' if self._status == 'show' else 'show'

        if self._status == 'hide':
            self.slider.hide()
            self.editValue.hide()
            self._update_label_text()
        else:
            self.slider.show()
            self.editValue.show()
            self.label.setText(self.name)

    def _update_label_text(self, *args):
        if self._status == 'hide':
            new_text = f'{self.name}{" = " if self.name else ""}{self._get_str_value()}'
            self.label.setText(new_text)

    def _get_str_value(self):
        if self._decimals:
            return f'{self.slider.value():.{self._decimals}f}'
        else:
            return str(int(self.slider.value()))

    def set_value(self, value: float, change_bounds: bool = False):
        emit_value = False
        if value < self.slider.minimum():
            if change_bounds:
                self.slider.setMinimum(value)
            else:
                emit_value = True
        if value > self.slider.maximum():
            if change_bounds:
                self.slider.setMaximum(value)
            else:
                emit_value = True
        self.slider.setValue(value)
        self.editValue.setText(self._get_str_value())
        if emit_value:
            self.slider.emit_value()

    def __init_ui__(self, orientation):
        layout = QHBoxLayout(self)

        self.slider = DoubleSlider(orientation, self,
                                   decimals=self._decimals)

        self.slider.setRange(*self.bounds)
        self.slider.setValue(self.init_value)
        self.slider.valueChangedByHand.connect(
            lambda *x: self.valueChanged.emit(self.slider.value()))
        self.editValue = QLineEdit(self._get_str_value())
        self.editValue.setMaximumWidth(self._EditMaximumWidth)
        self.editValue.setStyleSheet('QLineEdit {  border: none; }')
        self.label = QLabel(self.name)
        self.slider.valueChanged.connect(self._update_label_text)

        if not self._decimals:
            self.editValue.setValidator(QIntValidator(*self.bounds))

        self.slider.valueChangedByHand.connect(
            lambda _: self.editValue.setText(self._get_str_value()))
        self.editValue.editingFinished.connect(self._set_slider_value)

        layout.addWidget(self.label)
        layout.addWidget(self.slider)
        layout.addWidget(self.editValue)

        self.setFixedHeight(self._Height)
        min_width = self.fontMetrics().width(
            ' = '.join([self.name, self._get_str_value()])) + self._padding
        self.setMinimumWidth(min_width)

    def _set_slider_value(self):
        value = self.editValue.text()
        try:
            if self._decimals:
                value = float(value)
            else:
                value = int(value)
            self.slider.setValue(value)
            if value < self.bounds[0] or value > self.bounds[1]:
                raise ValueError()
        except ValueError:
            self.editValue.setText(self._get_str_value())
            color_animation(self.editValue)
        self.valueChanged.emit(self.slider.value())


class ControlSlider(AnimatedSlider):
    def __init__(self, *args, **kwargs):
        AnimatedSlider.__init__(self, *args, **kwargs)
        self._fixed = False

    def set_fixed(self):
        self.hide_and_show('hide')
        self._fixed = True
        self.context_menu_enabled = False
        self.label.setStyleSheet('QLabel {color: green; }')

    def set_unfixed(self):
        self._fixed = False
        self.context_menu_enabled = True
        self.label.setStyleSheet('QLabel {color: black; }')

    def on_clicked(self):
        if not self._fixed:
            AnimatedSlider.on_clicked(self)


class InfoButton(RoundedPushButton):
    def __init__(self, info, parent=None):
        RoundedPushButton.__init__(self, parent, icon=Icon('info'))
        self.setToolTip(info)


class AbstractInputParametersWidget(QWidget):
    class InputParameters(NamedTuple):
        name: str
        label: str
        type: type
        info: str = None
        none: str = False

    @property
    @abstractmethod
    def PARAMETER_TYPES(self):
        pass

    @property
    @abstractmethod
    def NAME(self):
        pass

    def __init__(self, parent=None):
        super(AbstractInputParametersWidget, self).__init__(parent)
        self.setWindowIcon(Icon('setup'))
        self.setWindowTitle(self.NAME)
        self.default_dict = self.dict_from_config()

    def get_parameters_dict(self):
        parameters_dict = dict()
        for p in self.PARAMETER_TYPES:
            value = getattr(self, p.name)
            if not p.none and value is None:
                return
            parameters_dict[p.name] = value
        return parameters_dict

    def dict_from_config(self):
        return read_config(self.NAME)

    def save_to_config(self, parameters_dict: dict):
        save_config(self.NAME, parameters_dict)

    def _get_layout(self,
                    input_parameter: 'InputParameters'):
        current_value = self.default_dict.get(input_parameter.name, '')
        if current_value is None:
            current_value = ''
        label_widget = QLabel(input_parameter.label)
        input_widget = QLineEdit(str(current_value))
        if input_parameter.type is int:
            input_widget.setValidator(QIntValidator())
        layout = QHBoxLayout()
        layout.addWidget(label_widget, Qt.AlignHCenter)
        layout.addWidget(input_widget, Qt.AlignHCenter)
        if input_parameter.info:
            info_button = InfoButton(input_parameter.info)
            layout.addWidget(info_button, Qt.AlignLeft)

        def get_input(s):
            return validate_scientific_value(input_widget, input_parameter.type, input_parameter.none)

        setattr(AbstractInputParametersWidget, input_parameter.name,
                property(get_input))
        return layout


class BasicInputParametersWidget(AbstractInputParametersWidget):
    apply_signal = pyqtSignal(dict)
    close_signal = pyqtSignal()

    def __init__(self, parent=None):
        AbstractInputParametersWidget.__init__(self, parent)
        self.form = QFormLayout()
        for p in self.PARAMETER_TYPES:
            self.form.addRow(self._get_layout(p))
        self.save_button = self.__init_save_button__()
        self.apply_button = self.__init_apply_button__()
        self.cancel_button = self.__init_cancel_button__()
        self.form.addWidget(self.save_button)
        self.form.addWidget(self.apply_button)
        self.form.addWidget(self.cancel_button)
        self.setLayout(self.form)
        center_widget(self)

    def __init_apply_button__(self):
        apply_button = QPushButton('Apply')

        apply_button.clicked.connect(self.on_apply)
        return apply_button

    def __init_cancel_button__(self):
        cancel_button = QPushButton('Cancel')
        cancel_button.clicked.connect(self.close)
        return cancel_button

    def on_apply(self, *args):
        parameters_dict = self.get_parameters_dict()
        if parameters_dict is not None:
            self.apply_signal.emit(parameters_dict)
            logger.debug(parameters_dict)
            if self.save_button.isChecked():
                self.save_to_config(parameters_dict)
            self.close()

    def close(self):
        self.close_signal.emit()
        AbstractInputParametersWidget.close(self)

    @staticmethod
    def __init_save_button__():
        save_button = QRadioButton('Save parameters')
        save_button.setChecked(True)
        return save_button



from typing import NamedTuple
from traceback import print_stack
import logging

import cv2

logger = logging.getLogger(__name__)


class _Mode(NamedTuple):
    name: str
    flag: int


INTERPOLATION_MODES = (
    _Mode('Nearest', cv2.INTER_NEAREST),
    _Mode('Bilinear', cv2.INTER_LINEAR),
    _Mode('Cubic', cv2.INTER_CUBIC),
    _Mode('Lanczos', cv2.INTER_LANCZOS4)
)


def get_mode(mode_name: str):
    # maybe a frozen dict would be a better solution, but it requires
    # additional dependencies
    for m in INTERPOLATION_MODES:
        if m.name == mode_name:
            return m
    else:
        logger.error(f'Unknown mode name {mode_name}. Traceback: \n {print_stack(limit=4)}')
        return INTERPOLATION_MODES[0]  # only for unexpected errors



from setuptools import setup, find_packages
from pathlib import Path


def read(filename: str):
    with open(Path(__file__).parent / filename, mode='r', encoding='utf-8') as f:
        return f.read()


setup(
    name='giwaxs_gui',
    packages=find_packages(),
    version='0.0.6',
    author='Vladimir Starostin',
    author_email='vladimir.starostin@uni-tuebingen.de',
    description='A graphical tool for basic analysis of GIWAXS images.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='GPLv3',
    python_requires='>=3.6.*',
    install_requires=[
        'numpy>=1.18.1',
        'opencv-python>=4.*.*.*',
        'scipy>=1.4.1',
        'h5py>=2.10.0',
        'PyQt5',
        'pyqtgraph'
    ],
    include_package_data=True,
    keywords='xray python giwaxs scientific-analysis',
    url='https://pypi.org/project/giwaxs-gui/',
)
